# coding=utf-8

from ClickSQL.factor_table.factortable import FactorTable

from test.conn import conn

ft = FactorTable(conn, cik_dt='datetime', cik_iid='code', strict_cik=False)

if __name__ == '__main__':
    ft.add_factor('EDGAR_LOG.edgar_log', factor_names=['size', 'ip', 'norefer'], )
    query_func = ft._node
    print(query_func('desc EDGAR_LOG.edgar_log'))
    # Node.add_factor('EDGAR_LOG.edgar_log', factor_names='cik', cik_dt='date', cik_iid='ip')
    # Node.add_factor('test.test2', factor_names='test2', cik_dt='cik_dt', cik_iid='cik_iid')
    # f = ft.fetch(reduced=True, add_limit=True)

    # ft2 = FactorTable(conn, cik_dt='datetime', cik_iid='code', strict_cik=False)
    # ft.set_cik_dt([datetime.datetime.now()])
    # res = ft.fetch(reduced=True, add_limit=True)

    # print(res)
    pass



# coding=utf-8
import re
from typing import Any, Union
from urllib.parse import unquote


class ArgumentError(Exception):
    """Raised when an invalid or conflicting function argument is supplied.

    This error generally corresponds to construction time state errors.

    """
    pass


def _rfc_1738_quote(text: str):
    return re.sub(r'[:@/]', lambda m: "%%%X" % ord(m.group(0)), text)


def _rfc_1738_unquote(text: str):
    return unquote(text)


def parse_rfc1738_args(name: str):
    """
    parse args and translate into dict by regex method
    ------------
    translate string format of
    'clickhouse://user:password@host:port/database'
    into
    dict version of it



    :param name: string
    :return: dict
    """
    pattern = re.compile(r'''
            (?P<name>[\w\+]+)://
            (?:
                (?P<user>[^:/]*)
                (?::(?P<password>.*))?
            @)?
            (?:
                (?:
                    \[(?P<ipv6host>[^/]+)\] |
                    (?P<ipv4host>[^/:]+)
                )?
                (?::(?P<port>[^/]*))?
            )?
            (?:/(?P<database>.*))?
            ''', re.X)

    m = pattern.match(name)
    if m is not None:
        components = m.groupdict()
        if components['database'] is not None:
            tokens = components['database'].split('?', 2)
            components['database'] = tokens[0]

        if components['user'] is not None:
            components['user'] = _rfc_1738_unquote(components['user'])

        if components['password'] is not None:
            components['password'] = _rfc_1738_unquote(components['password'])

        ipv4host: Union[str, Any] = components.pop('ipv4host')
        ipv6host: Union[str, Any] = components.pop('ipv6host')
        components['host'] = ipv4host or ipv6host

        if components['port'] is None:
            pass
        else:
            components['port'] = int(components['port'])

        return components
    else:
        raise ArgumentError(
            "Could not parse rfc1738 URL from string '%s'" % name)


if __name__ == '__main__':
    c = parse_rfc1738_args("clickhouse://test:sysy@199.199.199.199:1234/drre")
    print(c)
    pass



# coding=utf-8


from ClickSQL.clickhouse.ClickHouse import ClickHouseTableNode
from ClickSQL.clickhouse.ClickHouseExt import ClickHouseTableNodeExt
from ClickSQL.nodes.base import BaseSingleFactorTableNode

__version__ = '0.1.8.9.5'
__author__ = 'sn0wfree'
__all__ = ['ClickHouseTableNode', 'ClickHouseTableNodeExt', 'BaseSingleFactorTableNode']



# coding=utf8
import uuid


def uuid_hash(strings, uuid_type='uuid3', namespace=uuid.NAMESPACE_DNS):
    """
    :param strings:
    :param uuid_type:
    :param namespace:
    :return:
    """
    return getattr(uuid, uuid_type)(namespace, strings).hex



# coding=utf-8
import unittest

from ClickSQL.nodes.base import BaseSingleFactorTableNode
from test.conn import conn


class MyTestCaseDelayTasks(unittest.TestCase):
    def test_DelayTasks(self):
        FT = BaseSingleFactorTableNode(conn)
        FT('show tables', delay=True)

        self.assertEqual(len(FT.delay_tasks), 1)
        del FT

    def test_DelayTasks2(self):
        FT = BaseSingleFactorTableNode(conn)
        FT('show tables', delay=True)
        FT('show tables from system', delay=True)

        self.assertEqual(len(FT.delay_tasks), 2)
        del FT

    def test_DelayTasks3_delay_False(self):
        FT = BaseSingleFactorTableNode(conn)
        res = FT('show tables', delay=False)
        FT('show tables', delay=True)
        res2 = FT.delay_tasks.run(no_yield=True)[0]

        self.assertEqual(len(FT.delay_tasks), 1)
        self.assertEqual(all(res == res2), True)

    def test_DelayTasks4_delay_False(self):
        FT = BaseSingleFactorTableNode(conn)

        FT('show tables', delay=True)
        res = FT('show tables', delay=False)
        res2 = FT.delay_tasks.run(no_yield=True)[0]

        self.assertEqual(len(FT.delay_tasks), 1)
        self.assertEqual(all(res == res2), True)

    def test_delay_run(self):
        FT = BaseSingleFactorTableNode(conn)
        FT('show tables', delay=True)
        FT('show tables from system', delay=True)

        res = FT.delay_tasks.run(no_yield=True)

        self.assertEqual(len(res), 2)

    def test_delay_run_yield(self):
        FT = BaseSingleFactorTableNode(conn)
        FT('show tables', delay=True)
        FT('show tables from system', delay=True)

        res = FT.delay_tasks.run(no_yield=False)

        from collections import Iterable
        self.assertIsInstance(res, Iterable)

        # self.assertEqual(len(res), 2)

    def test_delay_run2(self):
        FT = BaseSingleFactorTableNode(conn)
        FT('show tables', delay=True)
        FT('show tables from system', delay=True)

        res = FT.delay_tasks.run(no_yield=True)

        self.assertEqual(True, all(res[0] == FT('show tables')))
        self.assertEqual(True, all(res[1] == FT('show tables from system')))


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
import pandas as pd
import re
from collections import namedtuple, ChainMap
from ClickSQL.clickhouse.ClickHouseCreate import TableEngineCreator

from ClickSQL.errors import ClickHouseTableExistsError, ParameterTypeError

factor_parameters = ('dt', 'code', 'value', 'fid')
ft_node = namedtuple('factortable', factor_parameters)


class ClickHouseTableNodeExt(TableEngineCreator):

    def __init__(self, conn_str: (str, dict, None) = None, **kwarg):
        super(ClickHouseTableNodeExt, self).__init__(conn_str=conn_str, **kwarg)
        self._src = conn_str
        self.db_table = self._para.database

    @staticmethod
    def __extend_dict_value__(conditions: (dict, ChainMap)):
        for s in conditions.values():
            if isinstance(s, str):
                yield s
            elif isinstance(s, (tuple, list)):
                for s_sub in s:
                    yield s_sub
            else:
                raise ValueError('filter settings get wrong type! only accept string and tuple of string')

    def explain(self, sql: str):
        return self.query(f'explain {sql}')

    @staticmethod
    def __obtain_other_filter__(other_filters):
        exits_keys = []
        for k, v in other_filters.items():
            if k in exits_keys:
                raise ValueError(f'found duplicated key: {k}')
            exits_keys.append(k)
            if isinstance(v, dict):
                yield v
            elif isinstance(v, (str, tuple)):
                yield {k: v}
            else:
                raise ValueError('filter settings get wrong type! only accept string and tuple of string')

    def get_sql(self, db_table: str, cols: (tuple, None, list) = None,
                order_by_cols: (list, tuple, None) = None,
                data_filter: dict = {}, include_filter=True,
                limit: (None, int, str) = None,
                **other_filters):
        """

        :param limit:
        :param db_table:
        :param data_filter:
        :param cols:
        :param include_filter:
        :param other_filters:
        :param order_by_cols: ['test1 asc','test2 desc']
        :return:
        """
        if cols is None:
            cols = ['*']
        elif len(cols) == 0:
            cols = ['*']

        if '*' in cols:
            cols = list(cols)
            # replace * into columns
            cols.pop(cols.index('*'))
            cols.extend(self.query(f"desc {db_table}")['name'].values.tolist())
        conditions = ChainMap(data_filter, *list(self.__obtain_other_filter__(other_filters)))
        filter_yield = self.__extend_dict_value__(conditions)
        if include_filter:
            cols = sorted(set(list(cols) + list(conditions.keys())))
        else:
            cols = sorted(set(cols))
        if order_by_cols is None:
            order_by_clause = ''
        elif len(order_by_cols) > 1:
            order_by_clause = f" order by ({','.join(order_by_cols)})"
        elif len(order_by_cols) == 1:
            order_by_clause = f" order by {','.join(order_by_cols)}"
        else:
            raise ValueError('order_by_cols get wrong length')
        if limit is None:
            limit_clause = ''
        elif isinstance(limit, int):
            limit_clause = f"limit {limit}"
        elif isinstance(limit, str) and limit.strip(' ').lower().startswith('limit'):
            limit_clause = limit
        else:
            raise ValueError(f'limit parameter got wrong type! only accept str,int or None, but got {type(limit)}')
        where_clause = ' and '.join(sorted(set(['1'] + list(filter_yield))))
        sql = f"select {','.join(cols)} from {db_table} where {where_clause} {order_by_clause} {limit_clause} "
        return sql

    def _execute(self, sql: str, **kwargs):
        return self.query(sql, **kwargs)
        # self.__execute__ = self.operator.query

    # @staticmethod
    # def _check_end_with_limit(string, pattern=r'[\s]+limit[\s]+[0-9]+$'):
    #     m = re.findall(pattern, string)
    #     if m is None or m == []:
    #         return False
    #     else:
    #         return True


if __name__ == '__main__':
    pass



# coding=utf-8
import unittest
import pandas as pd
import numpy as np
from ClickSQL.factor_test.factor_tools import detect_series_na


def generate_random_data(size=(100, 2), add_na=0):
    s = np.random.random(size=size)

    df = pd.DataFrame(s, columns=[f't{i}' for i in range(size[1])])
    na = pd.DataFrame([[None] * size[1]] * add_na, columns=[f't{i}' for i in range(size[1])])

    return df.append(na)


class MyTestCase_prepare(unittest.TestCase):
    def test_detect_series_na_nona(self):
        df = generate_random_data(size=(100, 1), add_na=0)
        na_value = detect_series_na(df['t0'])
        self.assertEqual(na_value, None)

    def test_detect_series_na_1na_noraise(self):
        df = generate_random_data(size=(100, 1), add_na=1)
        na_value = detect_series_na(df['t0'], raise_error=False)
        self.assertEqual(na_value, 1)

    def test_detect_series_na_2na_raise(self):
        df = generate_random_data(size=(100, 1), add_na=2)
        with self.assertRaises(ValueError):
            na_value = detect_series_na(df['t0'], raise_error=True)
            # self.assertEqual(na_value, 2)


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
import datetime
import sys
import time


def progress_test(counts, lenfile, speed):
    bar_length = 20
    w = (lenfile - counts) * speed
    eta = time.time() + w
    precent = counts / float(lenfile)

    ETA = datetime.datetime.fromtimestamp(eta)
    hashes = '#' * int(precent * bar_length)
    spaces = ' ' * (bar_length - len(hashes))
    sys.stdout.write("""\r%d%%|%s|read %d projects|Speed : %.4f |ETA: %s """ % (
        precent * 100, hashes + spaces, counts, speed, ETA))

    # sys.stdout.write("\rthis spider has already read %d projects, speed: %.4f/projects" % (counts,f2-f1))

    # sys.stdout.write("\rPercent: [%s] %d%%,remaining time: %.4f mins"%(hashes + spaces,precent,w))
    sys.stdout.flush()


def process_bar(iterable_obj, counts=None):
    if hasattr(iterable_obj, '__len__'):
        counts = len(iterable_obj)
    elif isinstance(counts, int):
        pass  # counts = counts
    else:
        iterable_obj = list(iterable_obj)
        counts = len(iterable_obj)
    for count, i in enumerate(iterable_obj):
        f = time.time()
        yield i
        progress_test(count, counts, time.time() - f)


if __name__ == '__main__':
    s = range(1000)
    for si in process_bar(s):
        pass
    pass



# coding=utf-8

from ClickSQL.clickhouse.ClickHouse import ClickHouseTableNode

from ClickSQL.errors import DatabaseNotExists


class Dumper(ClickHouseTableNode):
    def __init__(self, *args, **kwargs):
        super(Dumper, self).__init__(*args, **kwargs)

    def _dump_db_structure(self, *dbs: str):
        """

        :param dbs:
        :return:
        """
        res = filter(lambda x: x not in self.databases, dbs)
        for db in res:
            raise DatabaseNotExists(f'{db} is not exists')
        create_db_sql_list = [(f"CREATE DATABASE IF NOT EXISTS {db};\n\n", db) for db in dbs]
        return create_db_sql_list

    def _dump_table_structure(self, db: str, table=None):
        """

        :param db:
        :param table:
        :return:
        """
        h = []
        if table is None:
            tables = self.query(f"show tables from {db}")
            if tables.empty:
                h.append('\n\n')
            else:
                tables = tables.values.ravel().tolist()
                for table in tables:
                    sql = f"show create table {db}.{table}"
                    create_sql = self.query(sql)['statement'].values.ravel().tolist()[0] + ';\n\n'
                    h.append(create_sql)
        else:
            if isinstance(table, str):
                sql = f"show create table {db}.{table}"
                create_sql = self.query(sql)['statement'].values.ravel().tolist()[0] + ';\n\n'
                h.append(create_sql)
            else:
                raise ValueError('table parameter get wrong type!')
        return h

    def _dump_structure(self, *dbs, outfile='create_structure.sql'):
        """

        :param dbs:
        :param outfile:
        :return:
        """
        with open(outfile, 'w+') as f:
            for create_db_sql, db in self._dump_db_structure(*dbs):
                sql_list = [create_db_sql] + self._dump_table_structure(db)
                for sql in sql_list:
                    f.write(sql)

    def dump(self, *dbs, outfile='create_structure.sql', structure_only=True):
        """

        :param dbs:
        :param outfile:
        :param structure_only:
        :return:
        """
        if structure_only:
            self._dump_structure(*dbs, outfile=outfile)
        else:
            raise ValueError('structure_only only accept True, False is not available yet!')


if __name__ == '__main__':
    conn = "clickhouse://default:Imsn0wfree@47.104.186.157:8123/system"
    c = Dumper(conn)._dump_db_structure('system', 'raw')
    print(c)
    pass



# coding=utf-8
from multiprocessing import Process
from threading import Thread
from socket import *
import select


# sendfile 零拷贝
def recv_data(new_socket, client_info):
    print("客户端{}已经连接".format(client_info))
    # 接受数据
    raw_data = new_socket.recv(1024)
    while raw_data:
        print(f"收到来自{client_info}的数据：{raw_data}")
        raw_data = new_socket.recv(1024)
    new_socket.close()


# 多进程并发阻塞
def server_process():
    # 实例化socket对象
    socket_server = socket(AF_INET, SOCK_STREAM)
    # 设置端口复用
    socket_server.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    # 绑定IP地址和端口
    socket_server.bind(("", 7788))
    # 改主动为被动，监听客户端
    socket_server.listen(5)
    while True:
        # 等待连接
        new_socket, client_info = socket_server.accept()
        p = Process(target=recv_data, args=(new_socket, client_info))
        p.start()
        # 多进程会复制父进程的内存空间，所以父进程中new_socket也必须关闭
        new_socket.close()


# 多线程并发阻塞
def server_thread():
    # 实例化socket对象
    socket_server = socket(AF_INET, SOCK_STREAM)
    # 设置端口复用
    socket_server.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    # 绑定IP地址和端口
    socket_server.bind(("", 7788))
    # 改主动为被动，监听客户端
    socket_server.listen(5)
    while True:
        # 等待连接
        new_socket, client_info = socket_server.accept()
        p = Thread(target=recv_data, args=(new_socket, client_info))
        p.start()
        # 多线程共享一片内存区域，所以这里不用关闭
        # new_socket.close()


# 多路复用IO---select模型
def server_select():
    # 实例化对象
    socket_server = socket(AF_INET, SOCK_STREAM)
    # 绑定IP和端口
    socket_server.bind(("", 7788))
    # 将主动模式改为被动模式
    socket_server.listen(5)
    # 创建套接字列表
    socket_lists = [socket_server]
    # 等待客户端连接
    while True:
        # 只监听读的状态，程序阻塞在这，不消耗CPU，如果列表里面的值读状态变化后，就解阻塞
        read_lists, _, _ = select.select(socket_lists, [], [])
        # 循环有变化的套接字
        for sock in read_lists:
            # 判断是否是主套接字
            if sock == socket_server:
                # 获取新连接
                new_socket, client_info = socket_server.accept()
                print(f"客户端：{client_info}已连接")
                # 添加到监听列表中
                socket_lists.append(new_socket)
            else:
                # 不是主客户端，即接收消息
                raw_data = sock.recv(1024)
                if raw_data:
                    print(f"接收数据：{raw_data.decode('gb2312')}")
                else:
                    # 如果没有数据，则客户端断开连接
                    sock.close()
                    # 从监听列表中删除该套接字
                    socket_lists.remove(sock)


# 多路复用IO---epoll模型 参考https://blog.csdn.net/pysense/article/details/103840680
# 该代码无法在windows上运行，因为epoll是Linux2.6内核增加的新功能，windows并不支持。
def server_epoll():
    # 创建socket对象
    sock_server = socket(AF_INET, SOCK_STREAM)
    # 绑定IP和端口
    sock_server.bind(("", 7788))
    # 将主动模式设置为被动模式，监听连接
    sock_server.listen(5)
    # 创建epoll监测对象（Only supported on Linux 2.5.44 and newer.）
    epoll = select.epoll()
    # print("未注册epoll对象：{}".format(epoll))
    # 注册主套接字,监控读状态
    epoll.register(sock_server.fileno(), select.EPOLLIN)
    # print("注册了主套接字后：{}".format(epoll))
    # 创建字典，保存套接字对象
    sock_dicts = {}
    # 创建字典，保存客户端信息
    client_dicts = {}
    while True:
        # print("所有套接字：{}".format(sock_dicts))
        # print("所有客户端信息：{}".format(client_dicts))
        # 程序阻塞在这，返回文件描述符有变化的对象
        poll_list = epoll.poll()
        # print("有变化的套接字：{}".format(poll_list))
        for sock_fileno, events in poll_list:
            # print("文件描述符：{}，事件：{}".format(sock_fileno, events))
            # 判断是否是主套接字
            if sock_fileno == sock_server.fileno():
                # 创建新套接字
                new_sock, client_info = sock_server.accept()
                print(f"客户端：{client_info}已连接")
                # 注册到epoll监测中
                epoll.register(new_sock.fileno(), select.EPOLLIN)
                # 添加到套接字字典当中
                sock_dicts[new_sock.fileno()] = new_sock
                client_dicts[new_sock.fileno()] = client_info
            else:
                # 接收消息
                raw_data = sock_dicts[sock_fileno].recv(1024)
                if raw_data:
                    print(f"来自{client_dicts[sock_fileno]}的数据：{raw_data.decode('gb2312')}")
                else:
                    # 关闭连接
                    sock_dicts[sock_fileno].close()
                    # 注销epoll监测对象
                    epoll.unregister(sock_fileno)
                    # 数据为空，则客户端断开连接，删除相关数据
                    del sock_dicts[sock_fileno]
                    del client_dicts[sock_fileno]


if __name__ == '__main__':
    server_process()



# coding=utf-8
import unittest
import pandas as pd
import numpy as np
from ClickSQL.clickhouse.ClickHouseCreate import TableEngineCreator, ClickHouseTableExistsError

from test.conn import conn


class MyTestCase(unittest.TestCase):
    @staticmethod
    def connection():
        fdg = TableEngineCreator(conn_str=conn)

        # self.assertIsInstance(fdg, TableEngineCreator)
        return fdg

    def test_select_create_existed(self):
        fg = self.connection()
        with self.assertRaises(ClickHouseTableExistsError) as f:
            out = fg.create(db='test',
                            table='test3',
                            df_or_sql_or_dict='select * from test.test4 limit 1',
                            key_cols=['test1'],
                            engine_type='ReplacingMergeTree',
                            primary_key_cols=None,
                            sample_by_cols=None,
                            extra_format_dict=None,
                            partition_by_cols=None, execute=False)
            # print(out)
            # self.assertEqual(out, '1')

    def test_select_create_sql(self):
        fg = self.connection()
        out = fg.create(db='test',
                        table='test5',
                        df_or_sql_or_dict='select * from test.test4 limit 1',
                        key_cols=['test1'],
                        engine_type='ReplacingMergeTree',
                        primary_key_cols=None,
                        sample_by_cols=None,
                        extra_format_dict=None, check=False,
                        partition_by_cols=None, execute=False)

        res = 'CREATE TABLE IF NOT EXISTS test.test5   ENGINE = ReplacingMergeTree    ORDER BY  test1         SETTINGS index_granularity = 8192 as select * from test.test4 limit 1 '
        # print(out)
        self.assertEqual(out, res)

    def test_select_create_df(self):
        df = pd.DataFrame(np.random.random(size=(100, 2)), columns=['test1', 'test2'])
        fg = self.connection()
        out = fg.create(db='test',
                        table='test5',
                        df_or_sql_or_dict=df,
                        key_cols=['test1'],
                        engine_type='ReplacingMergeTree',
                        primary_key_cols=None,
                        sample_by_cols=None,
                        extra_format_dict=None, check=False,
                        partition_by_cols=None, execute=False)

        res = 'CREATE TABLE IF NOT EXISTS test.test5  ( `test1` Float64,`test2` Float64 ) ENGINE = ReplacingMergeTree   ORDER BY  test1         SETTINGS index_granularity = 8192  '
        # print(out)
        self.assertEqual(out, res)

    def test_select_create_dtypes(self):
        # df = pd.DataFrame(np.random.random(size=(100, 2)), columns=['test1', 'test2'])
        fg = self.connection()
        out = fg.create(db='test',
                        table='test5',
                        df_or_sql_or_dict={'test1': 'Float64', 'test2': 'Float64'},
                        key_cols=['test1'],
                        engine_type='ReplacingMergeTree',
                        primary_key_cols=None,
                        sample_by_cols=None,
                        extra_format_dict=None, check=False,
                        partition_by_cols=None, execute=False)

        res = 'CREATE TABLE IF NOT EXISTS test.test5  ( `test1` Float64,`test2` Float64 ) ENGINE = ReplacingMergeTree  ORDER BY  test1       SETTINGS index_granularity = 8192'
        # print(out)
        self.assertEqual(out, res)

    def test_select_create_sql_primary(self):
        # df = pd.DataFrame(np.random.random(size=(100, 2)), columns=['test1', 'test2'])
        fg = self.connection()
        out = fg.create(db='test',
                        table='test5',
                        df_or_sql_or_dict={'test1': 'Float64', 'test2': 'Float64'},
                        key_cols=['test1'],
                        engine_type='ReplacingMergeTree',
                        primary_key_cols=['test1'],
                        sample_by_cols=None,
                        extra_format_dict=None, check=False,
                        partition_by_cols=None, execute=False)

        res = 'CREATE TABLE IF NOT EXISTS test.test5  ( `test1` Float64,`test2` Float64 ) ENGINE = ReplacingMergeTree  ORDER BY  test1   PRIMARY KEY  test1      SETTINGS index_granularity = 8192'
        print(out)
        self.assertEqual(out, res)

    def test_select_create_df_primary(self):
        df = pd.DataFrame(np.random.random(size=(100, 2)), columns=['test1', 'test2'])
        fg = self.connection()
        out = fg.create(db='test',
                        table='test5',
                        df_or_sql_or_dict=df,
                        key_cols=['test1'],
                        engine_type='ReplacingMergeTree',
                        primary_key_cols=['test1'],
                        sample_by_cols=None,
                        extra_format_dict=None, check=False,
                        partition_by_cols=None, execute=False)

        res = 'CREATE TABLE IF NOT EXISTS test.test5  ( `test1` Float64,`test2` Float64 ) ENGINE = ReplacingMergeTree   ORDER BY  test1   PRIMARY KEY   test1        SETTINGS index_granularity = 8192  '
        # print(out)
        self.assertEqual(out, res)

    def test_select_create_dtypes_primary(self):
        # df = pd.DataFrame(np.random.random(size=(100, 2)), columns=['test1', 'test2'])
        fg = self.connection()
        out = fg.create(db='test',
                        table='test5',
                        df_or_sql_or_dict={'test1': 'Float64', 'test2': 'Float64'},
                        key_cols=['test1'],
                        engine_type='ReplacingMergeTree',
                        primary_key_cols=['test1'],
                        sample_by_cols=None,
                        extra_format_dict=None, check=False,
                        partition_by_cols=None, execute=False)

        res = 'CREATE TABLE IF NOT EXISTS test.test5  ( `test1` Float64,`test2` Float64 ) ENGINE = ReplacingMergeTree  ORDER BY  test1   PRIMARY KEY  test1      SETTINGS index_granularity = 8192'
        print(out)
        self.assertEqual(out, res)

    def test_select_create_dtypes_primary_partition(self):
        # df = pd.DataFrame(np.random.random(size=(100, 2)), columns=['test1', 'test2'])
        fg = self.connection()
        out = fg.create(db='test',
                        table='test5',
                        df_or_sql_or_dict={'test1': 'Float64', 'test2': 'Float64'},
                        key_cols=['test1'],
                        engine_type='ReplacingMergeTree',
                        primary_key_cols=['test1'],
                        sample_by_cols=None,
                        extra_format_dict=None, check=False,
                        partition_by_cols=['toString(test1)'], execute=False)

        res = 'CREATE TABLE IF NOT EXISTS test.test5  ( `test1` Float64,`test2` Float64 ) ENGINE = ReplacingMergeTree PARTITION BY  toString(test1)   ORDER BY  test1   PRIMARY KEY  test1      SETTINGS index_granularity = 8192'
        print(out)
        self.assertEqual(out, res)

    def test_create(self):
        fg = self.connection()
        out = fg.select(db_table='test.test', cols=['test1'],
                        sample=None,
                        array_join=None,
                        join=None,
                        prewhere=None,
                        where=None,
                        having=None,
                        group_by=None,
                        order_by=None,
                        limit_by=None,
                        limit=None)

        res = 'SELECT test1 FROM test.test            SETTINGS joined_subquery_requires_alias=0'
        # print(out)
        self.assertEqual(out.strip(), res.strip())

    def test_create_sample(self):
        print('sampling disable')
        fg = self.connection()
        out = fg.select(db_table='test.test', cols=['test1'],
                        sample=0.1,
                        array_join=None,
                        join=None,
                        prewhere=None,
                        where=None,
                        having=None,
                        group_by=None,
                        order_by=None,
                        limit_by=None,
                        limit=None)

        res = 'SELECT test1 FROM test.test            SETTINGS joined_subquery_requires_alias=0'
        # print(out)
        self.assertEqual(out.strip(), res.strip())

    def test_create_array_join(self):
        fg = self.connection()
        out = fg.select(db_table='test.test', cols=['g'],
                        sample=0.1,
                        array_join=['test1 as g'],
                        join=None,
                        prewhere=None,
                        where=None,
                        having=None,
                        group_by=None,
                        order_by=None,
                        limit_by=None,
                        limit=None)

        res = 'SELECT g FROM test.test  ARRAY JOIN test1 as g          SETTINGS joined_subquery_requires_alias=0'
        # print(out)
        self.assertEqual(out.strip(), res.strip())


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
import datetime
import hashlib
import os
import pickle
from collections import OrderedDict
from functools import wraps

__refresh__ = False
DEFAULT = './'
format_dict = {'Y': '%Y', 'm': "%Y-%m", 'd': "%Y-%m-%d",
               'H': '%Y-%m-%d %H', 'M': '%Y-%m-%d %H:%M', 'S': '%Y-%m-%d %H:%M:%S'}


def get_cache_path(enable_cache: bool = False):
    dt = datetime.datetime.today().strftime('%Y%m%d')
    # __cache_path__ == f"{default}"
    cache_path = os.path.join(DEFAULT, dt)
    if not os.path.exists(cache_path) and enable_cache:
        os.mkdir(cache_path)
    return cache_path


def date_format(granularity: str):
    if granularity in format_dict.keys():
        return format_dict.get(granularity)
    else:
        raise ValueError(f'date_format not support: {granularity}')
    pass


def prepare_args(func, arg, kwargs: dict, granularity: str = 'H', exploit_func_name: bool = True,
                 enable_cache: bool = False):
    """

    :param func:
    :param arg:
    :param kwargs:
    :param granularity:
    :param exploit_func_name:  cache file name whether include func name
    :param enable_cache:
    :return:
    """
    time_format_dimension = date_format(granularity)
    dt_str = datetime.datetime.now().strftime(time_format_dimension)
    kwargs = OrderedDict(sorted(kwargs.items(), key=lambda t: t[0]))  # sort kwargs to fix hashcode if sample input
    func_name = func.__name__.__str__()
    cls_obj = func.__qualname__ != func_name
    if cls_obj:
        obj = arg[0]
        obj = obj.__name__ if hasattr(obj, '__name__') else obj.__class__.__name__
        arg_tuple = tuple([obj] + list(map(str, arg[1:])))
    else:
        arg_tuple = arg
    key = pickle.dumps([func_name, arg_tuple, kwargs, dt_str])  # get the unique key for the same input
    if exploit_func_name:
        name = f"{func_name}_{hashlib.sha1(key).hexdigest()}_{dt_str}"  # create cache file name
    else:
        name = hashlib.sha1(key).hexdigest()  # create cache file name
    file_path = get_cache_path(enable_cache=enable_cache)
    return file_path, name


def write(fg, res):
    with open(fg, 'wb') as f:
        pickle.dump(res, f)


def read(fg):
    with open(fg, 'rb') as f:
        res = pickle.load(f)
    return res


def _cache(func, arg, kwargs, granularity='H', enable_cache: bool = False, exploit_func=True):
    """

    :param func:
    :param arg:
    :param kwargs:
    :param granularity:
    :param enable_cache:
    :param exploit_func:   cache file name whether include func name
    :return:
    """

    if enable_cache:
        file_path, name = prepare_args(func, arg, kwargs, granularity=granularity, exploit_func_name=exploit_func,
                                       enable_cache=enable_cache)
        fg = os.path.join(file_path, name)
        if os.path.exists(fg):
            return read(fg)
        else:
            res = func(*arg, **kwargs)
            write(fg, res)
            return res
    else:
        res = func(*arg, **kwargs)
        return res


def file_cache(**deco_arg_dict):
    # if callable(deco_arg_dict):
    #     @wraps(deco_arg_dict)
    #     def wrapped(*args, **kwargs):
    #         return _cache(deco_arg_dict, args, kwargs, granularity='d', enable_cache=False)
    #
    #     return wrapped
    # else:
    def _deco(func):
        @wraps(func)
        def __deco(*args, **kwargs):
            return _cache(func, args, kwargs, **deco_arg_dict)

        return __deco

    return _deco


if __name__ == '__main__':
    @file_cache(enable_cache=True)
    def test(a, b=2):
        return a, b


    class YGH(object):
        @staticmethod
        @file_cache(enable_cache=True)
        def test(a, b=2):
            return a, b

        @classmethod
        @file_cache(enable_cache=True)
        def test2(cls, a, b=3):
            return a, b

        @file_cache(enable_cache=True)
        def test3(self, a, b=3):
            return a, b


    print(test(1, b=3))

    print(YGH.test(1, b=3))
    print(YGH.test2(1, b=3))
    print(YGH().test3(1, b=3))
    pass



import unittest


from ClickSQL.clickhouse.ClickHouse import ClickHouseTableNode
from ClickSQL.clickhouse.ClickHouse import DatabaseTypeError
from ClickSQL.utils.parse_rfc_1738_args import ArgumentError


class MyTestConnectionCase(unittest.TestCase):
    # def test_connection_success(self):
    #     node = ClickHouseTableNode(conn)
    #     df1 = node.tables
    #
    #     self.assertGreater(len(df1), 5)

    def test_connection_settings_failure_empty(self):
        conn = ""
        with self.assertRaises(ArgumentError):
            node = ClickHouseTableNode(conn)

    def test_connection_settings_failure_wrong_port(self):
        conn = "clickhouse://default:121231@10.3.10.1/syste"
        with self.assertRaises(TypeError):
            node = ClickHouseTableNode(conn)

        conn = "clickhouse://default:121231@10.3.10.1:sd/syste"
        with self.assertRaises(ValueError):
            node = ClickHouseTableNode(conn)

        conn = "clickhouse://default:121231@10.3.10.1:/syste"
        with self.assertRaises(ValueError):
            node = ClickHouseTableNode(conn)

    def test_connection_settings_failure_wrong_host(self):
        conn = "clickhouse://default:121231@10.2.3.1:10/syste"
        with self.assertRaises(TypeError):
            node = ClickHouseTableNode(conn)

    def test_connection_settings_failure_wrong_symbol(self):
        conn = "mysql://default:121231@99.99.9.9:8123/system"

        with self.assertRaises(DatabaseTypeError):
            node = ClickHouseTableNode(conn)

    def test_connection_failure(self):
        conn = "clickhouse://default:121231@99.99.9.9:8123/system"
        node = ClickHouseTableNode(conn)
        df1 = node.tables

        self.assertGreater(len(df1), 5)



if __name__ == '__main__':
    unittest.main()



# coding=utf-8
import unittest

from ClickSQL import BaseSingleFactorTableNode
from test.conn import conn

import pandas as pd


class MyTestCaseClickHouseExt(unittest.TestCase):
    def node(self):
        conn_str = conn
        return BaseSingleFactorTableNode(conn_str)

    def test_mp(self):
        test = pd.read_excel('SH000001-1m.xlsx')
        test['code'] = '000001.sh'
        test['date'] = pd.to_datetime(test['date'])
        print('load data')

        Node = self.node()

        Node.create('test',
                    'test1',
                    test,
                    key_cols=['code', 'date'], execute=True, check=False)

        Node.insert_df(test, 'test', 'test1')


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
from ClickSQL.utils.cached_property import cached_property
from ClickSQL.utils.boost_up import boost_up
from ClickSQL.utils.file_cache import file_cache
from ClickSQL.utils.lazy_load import LazyInit
from ClickSQL.utils.parse_rfc_1738_args import parse_rfc1738_args
from ClickSQL.utils.singleton import singleton
from ClickSQL.utils.uuid_generator import uuid_hash

__all__ = ['cached_property', 'boost_up', 'file_cache', 'LazyInit', 'parse_rfc1738_args', 'singleton', 'uuid_hash']
if __name__ == '__main__':
    pass



# coding=utf-8
import asyncio
import gzip
import json
import warnings
from collections import namedtuple
from functools import partial, partialmethod
from urllib import parse

import numpy as np
import pandas as pd
import requests

from ClickSQL.conf import Config
from ClickSQL.errors import ParameterKeyError, ParameterTypeError, DatabaseTypeError, DatabaseError, \
    HeartbeatCheckFailure, ClickHouseTableNotExistsError,ServerError
from ClickSQL.utils import cached_property, file_cache, parse_rfc1738_args

"""
this will hold base function of clickhouse and it will apply a path of access clickhouse through clickhouse api service
this scripts will use none of clickhouse client and only depend on requests to make transactions with 
clickhouse-server

"""
if Config.get('ENGAGE_ASYNC', default=False):
    try:
        import nest_asyncio

        nest_asyncio.apply()  # allow run at jupyter and asyncio env
        ENGAGE_ASYNC = True
        from aiohttp import ClientSession
    except Exception as e:
        warnings.warn('Cannot run at jupyter or asyncio env! will use normal version instead!')
        ENGAGE_ASYNC = False
        from requests import Session as ClientSession
else:
    ENGAGE_ASYNC = False
    from requests import Session as ClientSession

node_parameters = ('host', 'port', 'user', 'password', 'database')
node = namedtuple('clickhouse', node_parameters)

PRINT_CHECK_RESULT = Config.get('PRINT_CHECK_RESULT', default=True)
GLOBAL_RAISE_ERROR = Config.get('GLOBAL_RAISE_ERROR', default=True)
SEMAPHORE = Config.get('SEMAPHORE', default=10)  # control async number for whole query list

available_queries_select = ('select', 'show', 'desc')
available_queries_insert = ('insert', 'optimize', 'create')
DEFAULT_CONNECT_SETTINGS = {'enable_http_compression': 1, 'send_progress_in_http_headers': 0,
                            'log_queries': 1, 'connect_timeout': 10, 'receive_timeout': 300,
                            'send_timeout': 300, 'output_format_json_quote_64bit_integers': 0,
                            'wait_end_of_query': 0}


# TODO change to queue mode change remove aiohttp depends
class ClickHouseHelper(object):
    @staticmethod
    def _check_df_and_dump(df: pd.DataFrame, describe_table: pd.DataFrame, auto_convert=True):
        """

        :param df:
        :param describe_table:
        :return:
        """
        describe_table = describe_table[~describe_table['default_type'].isin(['MATERIALIZED', 'ALIAS'])]
        non_nullable_columns = list(describe_table[~describe_table['type'].str.startswith('Nullable')]['name'])
        integer_columns = list(describe_table[describe_table['type'].str.contains('Int', regex=False)]['name'])

        if auto_convert:
            df = df.convert_dtypes()
        missing_in_df = {i: np.where(df[i].isnull(), 1, 0).sum() for i in non_nullable_columns}

        for i, value in missing_in_df.items():
            if value > 0:
                raise ValueError('"{0}" is not a nullable column, missing values are not allowed.'.format(i))

        df_columns = df.columns.tolist()
        each_row = df.to_dict(orient='records')

        for row in each_row:
            for col in df_columns:
                if pd.isnull(row[col]):
                    row[col] = None
                else:
                    if col in integer_columns:
                        try:
                            row[col] = int(row[col])
                        except Exception as e:
                            raise ValueError('Column "{0}" is {1}, while value "{2}"'.format(col,
                                                                                             describe_table[
                                                                                                 describe_table[
                                                                                                     'name'] == col].iloc[
                                                                                                 0]['type'], row[col]) + \
                                             ' in the dataframe column cannot be converted to Integer.')
            yield json.dumps(row, ensure_ascii=False)

    @staticmethod
    def _check_sql_type(sql: str) -> str:
        if sql.lower().startswith(available_queries_select):
            return 'select-liked'
        elif sql.lower().startswith(available_queries_insert):
            return 'insert-liked'

        else:
            raise ValueError('sql type cannot be supported! need upgrade!')

    @staticmethod
    def _transfer_sql_format(sql: str, convert_to: str, transfer_sql_format: bool = True) -> str:
        """
        provide a method which will translate a standard sql into clickhouse sql with might use format as suffix
        :param sql:
        :param convert_to:
        :param transfer_sql_format:
        :return:
        """
        if transfer_sql_format:
            if convert_to is None:
                clickhouse_format = 'JSON'
            else:
                clickhouse_format = 'JSONCompact' if convert_to.lower() == 'dataframe' else convert_to
            query_with_format = (sql.rstrip('; \n\t') + ' format ' + clickhouse_format).replace('\n', ' ').strip(' ')
            return query_with_format
        else:
            return sql

    @staticmethod
    def _load_into_pd(ret_value: (str, bytes), convert_to: str = 'dataframe', errors='ignore'):
        """
        will provide a approach to load data from clickhouse into pd.DataFrame format which may be easy to use

        :param ret_value:
        :param convert_to:
        :param errors:
        :return:
        """

        if convert_to.lower() == 'dataframe':
            result_dict = json.loads(ret_value, strict=False)
            meta = result_dict['meta']
            name = map(lambda x: x['name'], meta)
            df = pd.DataFrame.from_records(result_dict['data'], columns=list(name))
            for i in meta:
                if i['type'] in ['DateTime', 'Nullable(DateTime)']:  # translate format
                    # process datetime format
                    df[i['name']] = pd.to_datetime(df[i['name']], errors=errors)
            return df
        else:
            return ret_value

    @classmethod
    def _merge_settings(cls, settings: (None, dict), updated_settings: (None, dict) = None,
                        extra_settings: (None, dict) = None) -> dict:
        """

        :param settings:
        :param updated_settings:
        :return:
        """

        if updated_settings is None:
            updated_settings = {'enable_http_compression': 1, 'send_progress_in_http_headers': 0,
                                'log_queries': 1, 'connect_timeout': 10, 'receive_timeout': 300,
                                'send_timeout': 300, 'output_format_json_quote_64bit_integers': 0,
                                'wait_end_of_query': 0}
        elif not isinstance(updated_settings, dict):
            raise ParameterTypeError(f'updated_settings must be dict type, but get {type(updated_settings)}')
        else:
            pass
        if settings is not None and isinstance(settings, dict):
            invalid_setting_keys = set(settings.keys()) - set(updated_settings.keys())
            if len(invalid_setting_keys) > 0:
                raise ValueError('setting "{0}" are invalid, valid settings are: {1}'.format(
                    ','.join(invalid_setting_keys), ', '.join(updated_settings.keys())))
            updated_settings.update(settings)
        if extra_settings is not None and isinstance(extra_settings, dict):
            updated_settings.update(extra_settings)

        return {k: v * 1 if isinstance(v, bool) else v for k, v in updated_settings.items()}


class ClickHouseBaseNode(ClickHouseHelper):
    __slots__ = ('_db', '_connect_url', '_para', 'http_settings')  # 'max_async_query_once', 'is_closed'

    def __init__(self, **db_settings):
        """

        accepted_formats = ('DataFrame', 'TabSeparated', 'TabSeparatedRaw', 'TabSeparatedWithNames',
                        'TabSeparatedWithNamesAndTypes', 'CSV', 'CSVWithNames', 'Values', 'Vertical', 'JSON',
                        'JSONCompact', 'JSONEachRow', 'TSKV', 'Pretty', 'PrettyCompact',
                        'PrettyCompactMonoBlock', 'PrettyNoEscapes', 'PrettySpace', 'XML')

        :param db_settings:
        """
        self._check_db_settings_(db_settings, available_db_type=[node.__name__])
        self._para = node(db_settings['host'], db_settings['port'], db_settings['user'],
                          db_settings['password'], db_settings['database'])  # store connection information
        if '.' in self._para.database:
            self._db = self._para.database.split('.')[0]
        else:
            self._db = self._para.database
        self._connect_url = 'http://{user}:{passwd}@{host}:{port}'.format(user=self._para.user,
                                                                          passwd=self._para.password,
                                                                          host=self._para.host,
                                                                          port=self._para.port)
        self.http_settings = self._merge_settings(None, updated_settings=DEFAULT_CONNECT_SETTINGS,
                                                  extra_settings={'user': self._para.user,
                                                                  'password': self._para.password})
        _base_url = "http://{host}:{port}/?".format(host=self._para.host, port=int(self._para.port))
        self.__heartbeat_test__(_base_url)
        self._heartbeat_test_ = partialmethod(self.__heartbeat_test__, _base_url=_base_url)
        self.cache_query = partialmethod(self.execute, enable_cache=True, exploit_func=True)

    @staticmethod
    def _check_db_settings_(db_settings: dict, available_db_type=(node.__name__,)):
        """
        it is to check db setting whether is correct!
        :param db_settings:
        :return:
        """

        if isinstance(db_settings, dict):
            if db_settings['name'].lower() not in available_db_type:
                raise DatabaseTypeError(
                    f'database symbol is not accepted, now only accept: {",".join(available_db_type)}')

            missing_keys = filter(lambda x: x not in db_settings.keys(), node_parameters)  # can improve
            if len(tuple(missing_keys)) == 0:
                pass
            else:
                raise ParameterKeyError(f"the following keys are not at settings: {','.join(missing_keys)}")
        else:
            raise ParameterTypeError(f'db_setting must be dict type! but get {type(db_settings)}')

    @staticmethod
    def __heartbeat_test__(_base_url: str):
        """
        a function to test connection by normal way!

        alter function type into staticmethod
        :return:
        """

        ret_value = requests.get(_base_url)
        status_code = ret_value.status_code
        if status_code != 200:
            raise HeartbeatCheckFailure(f'heartbeat check failure at {_base_url} with status code: {status_code}')
        if PRINT_CHECK_RESULT:
            print('connection test: ', ret_value.text.strip())
        del ret_value

    def _post(self, url: str, sql: str, session, raise_error: bool = True):
        if self.http_settings['enable_http_compression'] == 1:
            with session.post(url, data=gzip.compress(sql.encode()),
                              headers={'Content-Encoding': 'gzip',
                                       'Accept-Encoding': 'gzip'}) as resp:
                result = resp.content
        else:
            with session.post(url, body=sql.encode(), ) as resp:
                result = resp.content

        # reason = resp.reason
        if resp.status_code != 200:
            if raise_error and GLOBAL_RAISE_ERROR:
                raise ServerError(result)
            else:
                # result = SmartBytes(result, resp.status_code)
                warnings.warn(str(result))
        return result

    async def _post_async(self, url: str, sql: str, session, raise_error: bool = True):
        """
        the async way to send post request to the server
        :param url:
        :param sql:
        :param session:
        :return:
        """
        if self.http_settings['enable_http_compression'] == 1:
            async with session.post(url, data=gzip.compress(sql.encode()),
                                    headers={'Content-Encoding': 'gzip',
                                             'Accept-Encoding': 'gzip'}) as resp:
                result = await resp.read()
        else:
            async with session.post(url, body=sql.encode(), ) as resp:
                result = await resp.read()

        # result = SmartBytes(result, resp.status)
        # reason = resp.reason
        if resp.status_code != 200:
            if raise_error and GLOBAL_RAISE_ERROR:
                raise DatabaseError(result)
            else:
                warnings.warn(str(result))
        return result

    def _compression_switched_request(self, query_with_format: (tuple, list, str), convert_to: str = 'dataframe',
                                      transfer_sql_format: bool = True, sem=None, raise_error=True):

        url = self._connect_url + '/?' + parse.urlencode(self.http_settings)
        transfer_sql = partial(self._transfer_sql_format, convert_to=convert_to,
                               transfer_sql_format=transfer_sql_format)
        with ClientSession() as session:
            if isinstance(query_with_format, str):
                result = self._post(url, transfer_sql(query_with_format), session,
                                    raise_error=raise_error)
            elif isinstance(query_with_format, (tuple, list)):
                result = [self._post(url, transfer_sql(sql), session, raise_error=raise_error) for sql
                          in query_with_format]
            else:
                raise ValueError('query_with_format must be str , list or tuple')
        return result

    async def _compression_switched_request_async(self, query_with_format: (tuple, list, str),
                                                  convert_to: str = 'dataframe',
                                                  transfer_sql_format: bool = True, sem=None, raise_error=True):
        """
        the core request operator with compression switch adaptor

        :param query_with_format:
        :param convert_to:
        :param transfer_sql_format:
        :param sem:
        :return:
        """
        url = self._connect_url + '/?' + parse.urlencode(self.http_settings)
        transfer_sql = partial(self._transfer_sql_format, convert_to=convert_to,
                               transfer_sql_format=transfer_sql_format)
        if sem is None:
            sem = asyncio.Semaphore(SEMAPHORE)  # limit async num
        async with sem:  # limit async number
            async with ClientSession() as session:
                if isinstance(query_with_format, str):
                    result = await self._post_async(url, transfer_sql(query_with_format), session,
                                                    raise_error=raise_error)
                elif isinstance(query_with_format, (tuple, list)):
                    result = [await self._post_async(url, transfer_sql(sql), session, raise_error=raise_error) for sql
                              in
                              query_with_format]
                else:
                    raise ValueError('query_with_format must be str , list or tuple')
        return result

    @classmethod
    def _load_into_pd_ext(cls, sql: (str, list, tuple), ret_value: (bytes, list, tuple), convert_to: str,
                          to_df: bool = True):
        """
        a way to parse into dataframe
        :param sql:
        :param ret_value:
        :param convert_to:
        :param to_df:
        :return:
        """
        if not to_df:
            result = ret_value
        elif isinstance(sql, str):
            # status code has been removed, if status code != 200 will raise error at post func !
            if ret_value != b'':
                result = cls._load_into_pd(ret_value, convert_to)
            else:
                result = ret_value
        elif isinstance(sql, (list, tuple)):
            result = [cls._load_into_pd(s, convert_to) if s != b'' else s for s in ret_value]
        else:
            raise ValueError(f'sql must be str or list or tuple,but get {type(sql)}')
        return result

    def get_describe_table(self, db, table, filter_type: (list, tuple, None) = None):
        """

        :param db:
        :param table:
        :param filter_type:
        :return:
        """
        if filter_type is None:
            filter_type = ['MATERIALIZED', 'ALIAS']

        describe_table = self.__execute__(f'describe table {db}.{table}', convert_to='dataframe',
                                          transfer_sql_format=True,
                                          loop=None, to_df=True, raise_error=True)
        return describe_table[
            ~describe_table['default_type'].isin(filter_type)]  # , integer_columns, non_nullable_columns

    def insert_df(self, df: pd.DataFrame, db: str, table: str, chunksize: int = 100000):
        """
        ## TODO need to test
        ## todo need test suitable chunksize
        :param chunksize:
        :param df:  data for inserted
        :param db: target database
        :param table:  target table
        :return:
        """
        describe_table = self.get_describe_table(db, table)
        dt_col = describe_table[describe_table['type'].isin(('DateTime', 'Nullable(DateTime)'))]['name'].values.ravel()
        for i in dt_col:
            df[i] = pd.to_datetime(df[i]).dt.strftime('%Y-%m-%d %H:%M:%S')
        row_count = df.shape[0]
        rows_data = (i for i in self._check_df_and_dump(df, describe_table))
        if row_count <= chunksize:
            query_with_format = ['insert into {0} format JSONEachRow \n{1}'.format('{}.{}'.format(db, table), '\n'.join(
                list(rows_data)))]
        else:
            from ClickSQL.utils.chunk import chunk
            query_with_format = ['insert into {0} format JSONEachRow \n{1}'.format('{}.{}'.format(db, table), '\n'.join(
                data)) for data in chunk(list(rows_data), chunksize)]
        self.__execute__(query_with_format, transfer_sql_format=False, loop=None, to_df=False, raise_error=True)

    def __execute__(self, sql: (str, list, tuple), convert_to: str = 'dataframe', transfer_sql_format: bool = True,
                    loop=None, to_df: bool = True, raise_error=True, async_mode=True):
        """
        the core execute function to run the whole requests and SQL or a list of SQL.
        :param sql: String or list or tuple
        :param convert_to:
        :param transfer_sql_format:
        :param loop:
        :param to_df:
        :return:
        """
        if async_mode and ENGAGE_ASYNC:
            sem = asyncio.Semaphore(SEMAPHORE)  # limit async num
            resp_list = self._compression_switched_request_async(sql, convert_to=convert_to,
                                                                 transfer_sql_format=transfer_sql_format, sem=sem,
                                                                 raise_error=raise_error)
            if loop is None:
                loop = asyncio.get_event_loop()  # init loop
            res = loop.run_until_complete(resp_list)
        else:
            res = self._compression_switched_request(sql, convert_to=convert_to,
                                                     transfer_sql_format=transfer_sql_format, sem=None,
                                                     raise_error=raise_error)
        result = self._load_into_pd_ext(sql, res, convert_to, to_df=to_df)
        return result

    def execute(self, *sql, convert_to: str = 'dataframe', loop=None, output_df: bool = True,
                enable_cache: bool = False, exploit_func: bool = True, raise_error: bool = True, async_mode=True):
        """
        execute sql or multi sql

        :param async_mode:
        :param raise_error:
        :param exploit_func:
        :param enable_cache:
        :param output_df:
        :param sql:
        :param convert_to:
        :param loop:
        :return:
        """

        func = file_cache(enable_cache=enable_cache, exploit_func=exploit_func)(self.__execute__)
        result = func(sql, convert_to=convert_to, transfer_sql_format=True, loop=loop,
                      to_df=True * output_df, raise_error=raise_error, async_mode=async_mode)
        return result[0] if len(sql) == 1 else result

    def __call__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return self.query(*args, **kwargs)

    def query(self, *sql: str, loop=None, output_df: bool = True, raise_error=True, async_mode=True):

        """
        add enable_cache and exploit_func

        ## TODO require to upgrade
        :param async_mode:
        :param raise_error:
        :param output_df:
        :param loop:
        :param sql:
        :return:
        """

        result = self.execute(*sql, convert_to='dataframe', loop=loop, output_df=output_df, enable_cache=False,
                              exploit_func=False, raise_error=raise_error, async_mode=async_mode)
        return result


class ClickHouseTableNode(ClickHouseBaseNode):
    __slots__ = ('_db', '_connect_url', '_para', 'http_settings', '_table')  # 'max_async_query_once', 'is_closed'

    @staticmethod
    def __parse_conn_str__(conn_str: (str, dict, None) = None, **kwargs):
        if conn_str is None:
            if kwargs != {}:
                db_settings = kwargs
                db_settings['name'] = node.__name__
            else:
                raise ParameterTypeError('database parameters cannot be parsed normally!')
        elif isinstance(conn_str, str):
            db_settings = parse_rfc1738_args(conn_str)
        elif isinstance(conn_str, dict):
            db_settings = conn_str
        else:
            raise ParameterTypeError(f'conn_str must be str or dict but get: {type(conn_str)}')
        return db_settings

    def __init__(self, conn_str: (str, dict, None) = None, **kwargs):
        """
        add kwargs to contain db settings
        :param conn_str:
        :param kwargs:
        """
        db_settings = self.__parse_conn_str__(conn_str=conn_str, **kwargs)
        if db_settings['port'] is None:  # add default port for clickhouse
            db_settings['port'] = 8123
        super(ClickHouseTableNode, self).__init__(**db_settings)
        self._table = self.tables[0]

    @property
    def columns(self):
        # db_table: str = None
        db_table = f"{self._db}.{self._table}"
        sql = f'desc {db_table}'
        res = self.execute(sql, convert_to='dataframe')['name'].tolist()
        return res

    @columns.setter
    def columns(self, table: str):
        if table in self.tables:
            self._table = table
        else:
            raise ClickHouseTableNotExistsError(f'{table} not at {self._db}')

    @property
    def tables(self):
        """
        show tables
        :return:
        """
        sql = 'SHOW TABLES FROM {db}'.format(db=self._db)
        res = self.execute(sql, convert_to='dataframe').values.ravel().tolist()
        return res

    @cached_property
    def databases(self):
        """
        show databases
        :return:
        """
        sql = 'SHOW DATABASES'
        res = self.execute(sql, convert_to='dataframe').values.ravel().tolist()
        return res

    @staticmethod
    def add_auto_increment_materialized_col(db_table: str, fid_col_name: str = 'fid'):
        """
        executable sql for clickhouse to add materialized column with auto-increment
        :param db_table:
        :param fid_col_name:
        :return:
        """

        exec_express = "bitOr(bitShiftLeft(toUInt64(now64()),24), rowNumberInAllBlocks())"
        return f"alter table {db_table} add column {fid_col_name} Int64  MATERIALIZED {exec_express}"

    def _check_exists(self, db_table: str, mode: str = 'table', output=True):
        """

        :param db_table:
        :param mode:
        :return:
        """
        # TODO check table exists
        if isinstance(db_table, str):
            if '.' in db_table:
                db, table = db_table.split('.')
            else:
                db, table = self._db, db_table
        else:
            raise ValueError('please input correct db.table information')
        if output:
            print(f'will detect {db}.{table} existence!')

        if mode == 'table':
            if db == self._db:
                return table in self.tables
            else:
                sql = f"show tables from {db}"
                tables = self.query(sql).values.ravel().tolist()
                return table in tables
        else:
            return db in self.databases


if __name__ == '__main__':
    pass



# coding=utf-8

def boost_up(func, tasks, core=None, method='Process', chunksize=None, star=True):
    if method == 'Process':
        from multiprocessing import Pool

    else:
        from multiprocessing.dummy import Pool

    pool = Pool(core)
    if star:
        h = pool.starmap(func, tasks, chunksize=chunksize)
    else:
        h = pool.map(func, tasks, chunksize=chunksize)
    pool.close()
    pool.join()
    return h


if __name__ == '__main__':
    pass



# coding=utf-8
import unittest

from ClickSQL import ClickHouseTableNodeExt
from test.conn import conn


class MyTestCaseClickHouseExt(unittest.TestCase):
    @staticmethod
    def connection():
        fdg = ClickHouseTableNodeExt(conn_str=conn)

        # self.assertIsInstance(fdg, TableEngineCreator)
        return fdg

    def test_get_sql(self):
        node = self.connection()

        cols = []

        data_filter = {}

        db_table = 'raw.v_st_dis_buy_info'
        include_filter = True
        limit = 'limit 10'
        order_by_cols = ['money asc']

        other_filters = {'money': 'money >= 1000000'}

        sql = node.get_sql(db_table, cols=cols,
                           order_by_cols=order_by_cols,
                           data_filter=data_filter, include_filter=include_filter,
                           limit=limit,
                           **other_filters)

        real_real = 'select biz_channel_orderid,channel_no,core_date,cr_dr_type,cust_no,end_time,in_account,m_five_account_num,maturity_date,merchant_no,money,money_type,option_type,order_code,order_term,out_account,product_id,product_name,return_flow,s_account_num,send_flow,settle_account_num,start_time,trans_rate,trans_status,trans_type from raw.v_st_dis_buy_info where 1 and money >= 1000000  order by money asc limit 10 '
        self.assertEqual(real_real, sql)

        print(1)


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
class GroupSQLUtils(object):
    @staticmethod
    def group_top(sql: str, by: (str, list, tuple), top: int = 5, cols: (str, None) = None):
        if isinstance(by, str):
            by = [by]
        if cols is None:
            cols = '*'
        gt_sql = f"select {cols} from ({sql})  limit {top} by {','.join(by)} "
        return gt_sql

    # group table
    @staticmethod
    def group_by(db_table_or_sql: str,
                 by: (str, list, tuple),
                 apply_func: (list,),
                 having: (list, tuple, None) = None):
        if isinstance(by, str):
            by = [by]
            group_by_clause = f"group by {by}"
        elif isinstance(by, (list, tuple)):
            group_by_clause = f"group by ({','.join(by)})"
        else:
            raise ValueError(f'by only accept str list tuple! but get {type(by)}')
        # db_table_or_sql = sql
        if having is None:
            having_clause = ''
        elif isinstance(having, (list, tuple)):
            having_clause = 'having ' + " and ".join(having)
        else:
            raise ValueError(f'having only accept list,tuple,None! but get {type(having)}')
        sql = f"select  {','.join(by + apply_func)}  from ({db_table_or_sql}) {group_by_clause} {having_clause} "
        # if execute:
        #     self.operator(sql)
        # else:
        return sql



import unittest

from ClickSQL.conf import Config


class MyTestCaseConfigs_get_set(unittest.TestCase):
    def test_Configs_int(self):
        Config.set('test', 1)
        print(Config.get('test'))
        self.assertIsInstance(Config.get('test'), int)

    def test_Configs_str(self):
        Config.set('test', '1')
        print(Config.get('test'))
        self.assertIsInstance(Config.get('test'), str)

    # def test_Configs_byte(self):
    #     Config.set('test', b'1')
    #     print(Config.get('test'))
    #     self.assertIsInstance(Config.get('test'), str)

    def test_Configs_tuple(self):
        Config.set('test', ('1', '2'))
        print(Config.get('test'))
        self.assertIsInstance(Config.get('test'), (list,))

    def test_Configs_list(self):
        Config.set('test', ['1', '2'])
        print(Config.get('test'))
        self.assertIsInstance(Config.get('test'), (list,))

    def test_Configs_dict(self):
        Config.set('test', {'1': 1, '2': 2})
        print(Config.get('test'))
        self.assertIsInstance(Config.get('test'), (dict,))

    # def test_Configs_set(self):
    #     Config.set('test', {'1', '2'})
    #     print(Config.get('test'))
    #     self.assertIsInstance(Config.get('test'), (set,))

    def test_Configs_bool(self):
        Config.set('test', True)
        print(Config.get('test'))
        self.assertIsInstance(Config.get('test'), bool)
        Config.set('test2', False)
        print(Config.get('test2'))
        self.assertIsInstance(Config.get('test2'), bool)


class MyTestCaseConfigs_getitem_setitem(unittest.TestCase):
    def test_Configs_int(self):
        Config['test'] = 1
        print(Config['test'])
        self.assertIsInstance(Config['test'], int)

    def test_Configs_str(self):
        Config['test'] = '1'
        print(Config['test'])
        self.assertIsInstance(Config['test'], str)

    # def test_Configs_byte(self):
    #     Config.set('test', b'1')
    #     print(Config.get('test'))
    #     self.assertIsInstance(Config.get('test'), str)

    def test_Configs_tuple(self):
        Config['test'] = ('1', '2')

        print(Config['test'])
        self.assertIsInstance(Config['test'], (list,))

    def test_Configs_list(self):
        Config['test'] = ['1', '2']
        print(Config['test'])
        self.assertIsInstance(Config['test'], (list,))

    def test_Configs_dict(self):
        Config['test'] = {'1': 1, '2': 2}
        # Config.set('test', {'1': 1, '2': 2})
        print(Config['test'])
        self.assertIsInstance(Config['test'], (dict,))

    # def test_Configs_set(self):
    #     Config.set('test', {'1', '2'})
    #     print(Config.get('test'))
    #     self.assertIsInstance(Config.get('test'), (set,))

    def test_Configs_bool(self):
        Config['test'] = True
        # Config.set('test', True)
        print(Config['test'])
        self.assertIsInstance(Config['test'], bool)
        Config['test2'] = False
        # Config.set('test2', False)
        print(Config['test2'])
        self.assertIsInstance(Config['test2'], bool)


class MyTestCaseConfigs_getattr_setattr(unittest.TestCase):
    def test_Configs_int(self):
        Config.test = 1
        print(Config.test)
        self.assertIsInstance(Config.test, int)

    def test_Configs_str(self):
        Config.test = '1'
        print(Config.test)
        self.assertIsInstance(Config.test, str)

    # def test_Configs_byte(self):
    #     Config.set('test', b'1')
    #     print(Config.get('test'))
    #     self.assertIsInstance(Config.get('test'), str)

    def test_Configs_tuple(self):
        Config.test = ('1', '2')

        print(Config.test)
        self.assertIsInstance(Config.test, (list,))

    def test_Configs_list(self):
        Config.test = ['1', '2']
        print(Config.test)
        self.assertIsInstance(Config.test, (list,))

    def test_Configs_dict(self):
        Config.test = {'1': 1, '2': 2}
        # Config.set('test', {'1': 1, '2': 2})
        print(Config.test)
        self.assertIsInstance(Config.test, (dict,))

    # def test_Configs_set(self):
    #     Config.set('test', {'1', '2'})
    #     print(Config.get('test'))
    #     self.assertIsInstance(Config.get('test'), (set,))

    def test_Configs_bool(self):
        Config.test = True
        # Config.set('test', True)
        print(Config.test)
        self.assertIsInstance(Config.test, bool)
        Config.test2 = False
        # Config.set('test2', False)
        print(Config.test2)
        self.assertIsInstance(Config.test2, bool)


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
import re

ERROR_PATTERNS = (
    # ClickHouse prior to v19.3.3
    re.compile(r'''Code:\ (?P<code>\d+),
            \ e\.displayText\(\)\ =\ (?P<type1>[^ \n]+):\ (?P<msg>.+?),
            \ e.what\(\)\ =\ (?P<type2>[^ \n]+)
        ''', re.VERBOSE | re.DOTALL),
    # ClickHouse v19.3.3+
    re.compile(r'''Code:\ (?P<code>\d+),
            \ e\.displayText\(\)\ =\ (?P<type1>[^ \n]+):\ (?P<msg>.+)
        ''', re.VERBOSE | re.DOTALL),
)


class ParameterKeyError(Exception):
    pass


class ParameterTypeError(Exception):
    pass


class DatabaseTypeError(Exception):
    pass


class DatabaseNotExists(Exception): pass


class DatabaseError(Exception):
    '''
    Raised when a database operation fails.
    '''
    pass


class ServerError(DatabaseError):
    """
    Raised when a server returns an error.
    """

    def __init__(self, message):
        self.code = None
        processed = self.get_error_code_msg(message.decode())
        if processed:
            self.code, self.message = processed
        else:
            # just skip custom init
            # if non-standard message format
            self.message = message
            super(ServerError, self).__init__(message)

    @classmethod
    def get_error_code_msg(cls, full_error_message):
        """
        Extract the code and message of the exception that clickhouse-server generated.
        See the list of error codes here:
        https://github.com/yandex/ClickHouse/blob/master/dbms/src/Common/ErrorCodes.cpp
        """
        for pattern in ERROR_PATTERNS:
            match = pattern.match(full_error_message)
            if match:
                # assert match.group('type1') == match.group('type2')
                return int(match.group('code')), match.group('msg').strip()

        return 0, full_error_message

    def __str__(self):
        if self.code is not None:
            return "{} ({})".format(self.message, self.code)


class HeartbeatCheckFailure(Exception):
    pass


class ClickHouseTableExistsError(Exception):
    pass


class ClickHouseTableNotExistsError(Exception):
    pass


class ArgumentError(Exception):
    """Raised when an invalid or conflicting function argument is supplied.

    This error generally corresponds to construction time state errors.

    """
    pass



# coding=utf-8
import os
import json
from ClickSQL.utils import singleton, uuid_hash

PREFIX = 'CONFIG:'


@singleton
class Conf(object):

    def __setattr__(self, key: str, value):
        self.set(key, value)

    def __setitem__(self, key: str, value):
        self.set(key, value)
        # os.environ[PREFIX + uuid_hash(key).upper()] = json.dumps(value)

    def __getattr__(self, item):
        return self.get(item)

    def __getitem__(self, item: str):
        return self.get(item)
        # result = os.getenv(PREFIX + uuid_hash(item).upper())
        # if result is None:
        #     return result
        # else:
        #     return json.loads(result)

    @staticmethod
    def get(key: str, default=None):
        """

        :param key:
        :param default:
        :return:
        """

        key_str = PREFIX + uuid_hash(key).upper()
        result = os.environ.get(key_str, default=None)
        if result is None:
            return default
        else:
            return json.loads(result)

    @staticmethod
    def set(key: str, value):

        key_str = PREFIX + uuid_hash(key).upper()

        os.environ[key_str] = json.dumps(value)

    @staticmethod
    def show_configs():
        return {k: os.getenv(k) for k in filter(lambda x: x.startswith(PREFIX) and len(x) == 39, os.environ.keys())}


Config = Conf()

if __name__ == '__main__':
    # C = Conf()
    # C.test = '1'
    #
    # print(C.show_config())

    pass



# coding=utf-8
import copy
import warnings
from collections import namedtuple, deque

from functools import wraps

from ClickSQL.clickhouse.ClickHouseExt import ClickHouseTableNodeExt
from ClickSQL.errors import ClickHouseTableNotExistsError
from ClickSQL.nodes.groupby import GroupSQLUtils
from ClickSQL.nodes.merge import MergeSQLUtils

complex_sql_select_count = 4
factor_parameters = ('dt', 'code', 'value', 'fid')
ft_node = namedtuple('factortable', factor_parameters)

# class FactorBackendCH(object):
#     __slots__ = ['_src', 'node', 'db_table']
#
#     def __init__(self, src: str):
#         """
#
#         :param src:  sample: clickhouse://test:sysy@199.199.199.199:1234/drre
#         """
#         self._src = src
#         self.node = ClickHouseTableNodeExt(conn_str=src)
#         self.db_table = self.node._para.database
#
#     def __call__(self, sql: str, **kwargs):
#         return self.node.query(sql, **kwargs)


"add auto-increment col by materialized bitOr(bitShiftLeft(toUInt64(now64()),24), rowNumberInAllBlocks()) "


class DelayTasks(deque):
    def __init__(self,_query, *args, **kwargs):
        super(DelayTasks, self).__init__(*args, **kwargs)

        self._query = _query

    def run(self, no_yield=False):
        result = (self._query(sql) for sql in self)
        if no_yield:
            return list(result)

        else:
            return result


class BaseSingleQueryBaseNode(object):
    """
    init operator
    and set fid_ck,dt_max_1st,execute,no_self_update for update functions
    kwargs is the conditions for select
    """
    __Name__ = "基础因子库单因子基类"
    __slots__ = (
        'operator', 'db', 'table', 'db_table', '_kwargs', '_raw_kwargs', 'status', '_INFO', 'depend_tables',
        '_fid_ck', '_dt_max_1st', '_execute', '_no_self_update', 'delay_tasks'
    )

    def __init__(self, src: str, db_table: (None, str) = None, info=None, **kwargs):
        """

        :type kwargs: object
        :param src: string sample: clickhouse://test:sysy@199.199.199.199:1234/drre
        :param db_table:
        :param info:
        :param kwargs:  data_filter will store operator for some cols:
                { cols: (tuple, None, list) = None,
                 order_by_cols: (list, tuple, None) = None,
                 data_filter: dict = {}, include_filter=True,
                 **other_filters}




        """

        self.operator = ClickHouseTableNodeExt(src)

        # self._execute = self._operator._execute

        if db_table is None:
            src_db_table = self.operator.db_table
            db_table_all, db, table = self._db_split(src_db_table)
            # if '.' in src_db_table:
            #     self.db_table = src_db_table
            # else:
            #     raise ValueError('db_table parameter get wrong type!')
        elif isinstance(db_table, str):
            db_table_all, db, table = self._db_split(db_table)
        else:
            raise ValueError('db_table only accept str!')
        self.db_table = db_table_all
        # db, table = self.db_table.split('.')
        self.db = db
        self.table = table
        # self.depend_tables = [self.db_table]
        self._kwargs = kwargs
        self._raw_kwargs = kwargs
        self.status = 'SQL'
        self._INFO = info

        self.delay_tasks = DelayTasks(self.operator)
        # self.delay_tasks._query = self.operator

    @staticmethod
    def _db_split(src_db_table):
        if isinstance(src_db_table, str):
            if '.' in src_db_table:
                db, table = src_db_table.split('.')
                return src_db_table, db, table
            else:
                return f"{src_db_table}.None", src_db_table, 'None'
        else:
            raise ValueError('db_table parameter get wrong type! only accept str')

    @wraps(ClickHouseTableNodeExt.insert_df)
    def insert_df(self, *args, **kwargs):
        self.operator.insert_df(*args, **kwargs)

    # create table
    @wraps(ClickHouseTableNodeExt.create)
    def create(self, *args, **kwargs):
        return self.operator.create(*args, **kwargs)

    def _update(self, **kwargs):
        """
        update kwargs settings
        :param kwargs:
        :return:
        """

        self._kwargs.update(kwargs)

    def __str__(self, *args, **kwargs):
        return self.__sql__

    def __len__(self) -> int:
        """
        Returns length of info axis, but here we use the index.
        """
        sql = f"select count(1) as rows from ({self.__sql__})"
        rows = self.operator(sql)['rows'].values[0]
        return rows

    @property
    def __sql__(self):
        return self.operator.get_sql(db_table=self.db_table, **self._kwargs)

    def decorate(self, target_conn_args='conn'):
        def afunc(func):
            def _afunc(*args, **kwargs):
                if target_conn_args in kwargs.keys():
                    if kwargs[target_conn_args] is None:
                        kwargs[target_conn_args] = self.operator
                    elif callable(kwargs[target_conn_args]):
                        pass
                    else:
                        raise KeyError(f'target variable:{target_conn_args} had been setup into other value! ')
                else:
                    raise KeyError(f'cannot find target variable:{target_conn_args} ')
                return func(*args, **kwargs)

            return _afunc

        return afunc

    def __call__(self, *sql, **kwargs):
        """

        add delay function to delay execute sql

        :param sql:
        :param kwargs:
        :return:
        """

        if 'delay' in kwargs.keys():
            if kwargs['delay']:
                if isinstance(sql, str):
                    self.delay_tasks.append(sql)
                else:
                    for s in sql:
                        self.delay_tasks.append(s)
            else:
                kwargs.pop('delay')
                return self.operator(*sql, **kwargs)
        else:

            return self.operator(*sql, **kwargs)

    @property
    def __factor_id__(self):  # add iid function get factor table id
        return hash(self.__sql__)

    def __getitem__(self, key: (list, str)):
        """
        directly execute
        :param key:
        :return:
        """
        if isinstance(key, list):
            sql = f"select {','.join(key)} from ({self.__sql__})"
            return self.operator(sql)
        elif isinstance(key, str):
            sql = f"select {key} from ({self.__sql__})"
            return self.operator(sql)
        else:
            raise ValueError('key only accept list or str')

    @property
    def __system_tables__(self):
        sql = f"select total_rows,engine from system.tables where database ='{self.db}' and name='{self.table}'"
        res = self.operator(sql)
        return res

    def _detect_complex_sql(self, warn=False):
        """
        detect complex sql
        :param warn:
        :return:
        """
        sql = self.__sql__.lower()
        if 'select' in sql and len(sql.split('select ')) >= complex_sql_select_count:
            if warn:
                warnings.warn('this sql may be a complex sql!')
            else:
                raise ValueError('this sql may be a complex sql!')

    @property
    def table_exist(self):
        """
        return table exists status
        :return:
        """

        return not self.__system_tables__.empty

    @property
    def table_engine(self):
        """
        return table engine
        :return:
        """
        if self.table_exist:
            return self.__system_tables__['engine'].values[0]
        else:
            raise ClickHouseTableNotExistsError(f'{self.db_table} is not exists!')

    def fetch(self, num=1000, pattern=r'[\s]+limit[\s]+[0-9]+$', **kwargs):
        """
        fetch first 1000 line
        :return:
        """
        sql = self.__sql__
        self._detect_complex_sql()
        end_with_limit = self.operator._check_end_with_limit(sql, pattern=pattern)
        if end_with_limit:
            return self.operator(sql, **kwargs)
        else:
            try:
                return self.operator(sql + f' limit {num}', **kwargs)
            except Exception as e:
                return self.operator(sql, **kwargs)

    def fetch_all(self, **kwargs):
        """
        fetch all data
        :return:
        """
        self._detect_complex_sql()
        return self.operator(self.__sql__, **kwargs)

    @property
    def row_count(self):
        return self.__len__()

    @property
    def shape(self):
        """
        shape like pandas

        :return:
        """

        return self.row_count, self.col_count

    @property
    def dtypes(self):
        sql = f"desc ({self.__sql__})"
        dtypes = self.operator(sql)
        return dtypes

    @property
    def columns(self):
        return self.dtypes['name'].values.tolist()

    @property
    def col_count(self):
        return self.dtypes.shape[0]

    @property
    def empty(self):
        return self.total_rows == 0

    @property
    def total_rows(self):
        """
        return row count
        :return:
        """

        # sql = f"-- select count(1) as row_count from {self.db_table}"
        temp = self.__system_tables__
        if not temp.empty:
            return temp['total_rows'].values[0]
        else:
            raise ClickHouseTableNotExistsError(f'{self.db_table} is not exists!')

    def nlargest(self, top: int, columns: list, execute: bool = True, extra_cols: (str, None) = None):
        """
        return largest n by columns
        :param extra_cols:
        :param top:
        :param columns:
        :param execute:
        :return:
        """
        by = [f'{c} desc' for c in columns]
        sql = GroupSQLUtils.group_top(self.__sql__, by=by, top=top, cols=extra_cols)
        if execute:
            return self.operator(sql)
        else:
            return sql

    def nsmallest(self, top: int, columns: list, execute: bool = True, extra_cols: (str, None) = None):
        """
        return largest n by columns
        :param extra_cols:
        :param top:
        :param columns:
        :param execute:
        :return:
        """
        by = [f'{c} asc' for c in columns]
        sql = GroupSQLUtils.group_top(self.__sql__, by=by, top=top, cols=extra_cols)
        if execute:
            return self.operator(sql)
        else:
            return sql

    def groupby(self, by: (str, list, tuple), apply_func: (list,), having: (list, tuple, None) = None, execute=True):
        """

        :param execute:
        :param by:
        :param apply_func:
        :param having:
        :return:
        """
        sql = GroupSQLUtils.group_by(self.__sql__, by=by, apply_func=apply_func, having=having)
        if execute:
            return self.operator(sql)
        else:
            return sql

    def merge(self, seconds, using: (list, str, tuple), join_type='all full join', cols: (list, str, None) = None,
              execute=True):
        """

        :param seconds:
        :param using:
        :param join_type:
        :param cols:
        :param execute:
        :return:
        """
        if seconds.lower().startswith('select'):
            pass
        else:
            seconds = f" select * from {seconds}"
        sql = MergeSQLUtils._merge(self, seconds, using=using, join_type=join_type, cols=cols)
        if execute:
            return self.operator(sql)
        else:
            return sql

    # update table
    # def __lshift__(self, src_db_table):
    #     print('lshift')
    #     fid_ck = self._fid_ck
    #     dt_max_1st = self._dt_max_1st
    #     execute = self._execute
    #     no_self_update = self._no_self_update
    #
    #     if isinstance(src_db_table, str):
    #         src_conn = copy.deepcopy(self.operator._src).replace(self.db_table, src_db_table)
    #         src_db_table = BaseSingleQueryBaseNode(src_conn, cols=['*'])
    #     elif isinstance(src_db_table, BaseSingleQueryBaseNode):
    #         pass
    #     else:
    #         raise ValueError('src_db_table is not valid! please check!')
    #
    #     if src_db_table.empty:
    #         raise ValueError(f'{src_db_table.db_table} is empty')
    #     # check two table are same
    #     if no_self_update and self.db_table == src_db_table.db_table and self.__factor_id__ == src_db_table.__factor_id__:
    #         dst = src_db_table.db_table
    #         src = self.db_table
    #         raise ValueError(
    #             f'Detect self-update process! these operator attempts to update data from {src} to {dst}')
    #
    #     update_status = 'full' if self.empty else 'incremental'
    #
    #     func = getattr(UpdateSQLUtils, f'{update_status}_update')
    #     sql = func(src_db_table, self.db_table, fid_ck, dt_max_1st=dt_max_1st)
    #     if execute:
    #         self.operator(sql)
    #     return sql, update_status
    #
    # # update table
    # def __rshift__(self, dst_db_table: str):
    #     """
    #
    #     UpdateSQLUtils
    #
    #     :param dst_db_table:
    #     :return:
    #     """
    #     # print('rshift')
    #     fid_ck = self._fid_ck
    #     dt_max_1st = self._dt_max_1st
    #     execute = self._execute
    #     no_self_update = self._no_self_update
    #     if self.empty:
    #         raise ValueError(f'{self.db_table} is empty')
    #
    #     if isinstance(dst_db_table, str):
    #         dst_conn = copy.deepcopy(self.operator._src)
    #         dst_db_table = BaseSingleQueryBaseNode(
    #             dst_conn.replace(self.db_table, dst_db_table),
    #             cols=['*']
    #         )
    #     elif isinstance(dst_db_table, BaseSingleQueryBaseNode):
    #         pass
    #     else:
    #         raise ValueError('dst_db_table is not valid! please check!')
    #     # check two table are same
    #     if no_self_update and self.db_table == dst_db_table.db_table:
    #         if self.__factor_id__ == dst_db_table.__factor_id__:
    #             dst = dst_db_table.db_table
    #             src = self.db_table
    #             raise ValueError(
    #                 f'Detect self-update process! these operator attempts to update data from {src} to {dst}')
    #
    #     update_status = 'full' if dst_db_table.empty else 'incremental'
    #
    #     func = getattr(UpdateSQLUtils, f'{update_status}_update')
    #     sql = func(self, dst_db_table, fid_ck, dt_max_1st=dt_max_1st)
    #     if execute:
    #         self.operator(sql)
    #     return sql, update_status


# class BaseSingleFactorNode(BaseSingleFactorBaseNode):
#     __slots__ = (
#         'operator', 'db', 'table', 'db_table', '_kwargs', '_raw_kwargs', 'status', '_INFO', 'depend_tables',
#         '_fid_ck', '_dt_max_1st', '_execute', '_no_self_update'
#     )
#
#     def __init__(self, *args, **kwargs):
#         super(BaseSingleFactorNode, self).__init__(*args, **kwargs)


# class UpdateSQLUtils(object):
#
#     @staticmethod
#     def full_update(src_db_table: BaseSingleQueryBaseNode, dst_db_table: BaseSingleQueryBaseNode, **kwargs):
#         # dst_db_table = dst_db_table.db_table
#         # dst_db, dst_table = dst_db_table.db, dst_db_table.table
#         dst_table_type = dst_db_table.table_engine
#         dst = dst_db_table.db_table
#         if dst_table_type == 'View':
#             raise ValueError(f'{dst} is View ! cannot be updated!')
#         insert_sql = f"insert into {dst} {src_db_table}"
#         return insert_sql
#
#     @staticmethod
#     def incremental_update(src_db_table: BaseSingleQueryBaseNode, dst_db_table: BaseSingleQueryBaseNode,
#                            fid_ck: str, dt_max_1st=True, inplace=False, **kwargs):
#         # src_db_table = src_table.db_table
#         # src_table_type = src_db_table.table_engine
#         dst_table_type = dst_db_table.table_engine
#         dst = dst_db_table.db_table
#         if dst_table_type == 'View':
#             raise ValueError(f'{dst} is View ! cannot be updated!')
#         if dt_max_1st:
#             order_asc = ' desc'
#         else:
#             order_asc = ' asc'
#         sql = f" select distinct {fid_ck} from {dst} order by {fid_ck} {order_asc} limit 1 "
#         fid_ck_values = src_db_table.operator(sql).values.ravel().tolist()[0]
#         if inplace:
#             src_db_table._update(**{f'{fid_ck} as src_{fid_ck}': f' {fid_ck} > {fid_ck_values}'})
#             insert_sql = f"insert into {dst} {src_db_table}"
#         else:
#             src_db_table_copy = copy.deepcopy(src_db_table)
#             src_db_table_copy._update(**{f'{fid_ck} as src_{fid_ck}': f' {fid_ck} > {fid_ck_values}'})
#             insert_sql = f"insert into {dst} {src_db_table_copy}"
#
#         return insert_sql


class BaseSingleFactorTableNode(BaseSingleQueryBaseNode):
    __slots__ = (
        'operator', 'db', 'table', 'db_table', '_kwargs', '_raw_kwargs', 'status', '_INFO', 'depend_tables',
        # '_fid_ck', '_dt_max_1st',
        '_execute', '_no_self_update', '_complex'
    )

    def __init__(self, src: str, db_table: (None, str) = None, info=None,
                 # fid_ck: str = 'fid', dt_max_1st: bool = True,
                 execute: bool = False,
                 # no_self_update: bool = True,
                 **kwargs):
        super(BaseSingleFactorTableNode, self).__init__(src, db_table=db_table, info=info, **kwargs)
        self._complex = False
        # self._fid_ck = fid_ck
        # self._dt_max_1st = dt_max_1st
        self._execute = execute
        # self._no_self_update = no_self_update

    # def execute(self, sql):
    #     ## add execute usage
    #     return self.operator(sql)
    #
    # def drop_table(self, target: str):
    #     if '.' not in target:
    #         raise ValueError('drop table must tell correspond database')
    #     self.operator(f'drop table if exists {target}')
    #
    # def drop_db(self, target: str):
    #     self.operator(f'drop database if exists {target}')


## https://zhuanlan.zhihu.com/p/297623539


if __name__ == '__main__':
    # v_st_dis_buy_info = BaseSingleFactorTableNode(
    #     'clickhouse://default:Imsn0wfree@47.104.186.157:8123/raw.v_st_dis_buy_info',
    #     cols=None,
    #     order_by_cols=['money asc'],
    #     money='money >= 1000000', limit='limit 10'
    # )
    # print(v_st_dis_buy_info)
    #
    # # factor >> 'test.test'
    # # print(factor)
    # c = v_st_dis_buy_info.fetch(1000)
    # print(c)
    # sql = v_st_dis_buy_info.merge('select cust_no, product_id, money as s from sample.sample where money >= 100000', using=['cust_no', 'product_id'])
    # print(sql)
    # c = factor('show tables from raw')
    # c2 = factor.groupby(['test2'], apply_func=['sum(fid)'])
    # print(c2)

    # print(1 >> 2)
    import numpy as np
    import pandas as pd

    data = np.random.random(size=(100, 2))
    res = pd.DataFrame(data, columns=['test1', 'test2'])
    res.groupby('test1')

    pass



# coding=utf-8
import pandas as pd
import re

from ClickSQL.clickhouse.ClickHouse import ClickHouseTableNode
from ClickSQL.errors import ClickHouseTableExistsError, ParameterTypeError
import warnings


class SQLBuilder(object):

    @staticmethod
    def _assemble_sample(sample=None) -> str:
        warnings.warn('currently sampling is not supported!')
        sample_clause = '' if sample is None else f'SAMPLE {sample}'
        return ''

    @staticmethod
    def _assemble_array_join(array_join_list: (list, tuple, None) = None) -> str:

        array_join_clause = '' if array_join_list is None else f"ARRAY JOIN {', '.join(array_join_list)}"

        return array_join_clause

    @staticmethod
    def _assemble_join(join_info_dict: (dict, None) = None) -> str:
        if join_info_dict is None:
            join_clause = ''
        elif isinstance(join_info_dict, dict):
            join_type = join_info_dict.get('type', None)
            sql = join_info_dict.get('sql', None)
            if join_type is None:
                raise ValueError('join_info_dict cannot locate join_type condition')
            if sql is None:
                raise ValueError('join_info_dict cannot locate sql clause')

            on_ = join_info_dict.get('ON')
            if on_ is None:
                using_ = join_info_dict.get('USING')
                if using_ is None:
                    raise ValueError('join_info_dict cannot locate ON or USING condition')
                else:
                    join_clause = f'{join_type} ({sql}) USING ({using_})'
            else:
                join_clause = f'{join_type} ({sql}) ON {on_}'
        else:
            raise ValueError('join_info_dict must accept dict or None')
        return join_clause

    @staticmethod
    def _assemble_where_like(a_list, prefix='WHERE') -> str:

        sample_clause = '' if a_list is None else f"{prefix} {' and '.join(a_list)}"

        return sample_clause

    @staticmethod
    def _assemble_group_by(group_by_cols=None) -> str:
        if group_by_cols is None:
            sample_clause = ''
        else:
            group_by_cols_str = ','.join(group_by_cols)
            sample_clause = f'GROUP BY ({group_by_cols_str})'
        return sample_clause

    @staticmethod
    def _assemble_order_by(order_by_cols=None) -> str:

        sample_clause = '' if order_by_cols is None else f"ORDER BY ({','.join(order_by_cols)})"

        return sample_clause

    @staticmethod
    def _assemble_limit_by(limit_n_by_dict=None) -> str:

        if limit_n_by_dict is None:
            sample_clause = ''
        else:
            N = limit_n_by_dict['N']
            order_by_cols_str = ','.join(limit_n_by_dict['limit_by_cols'])
            sample_clause = f'LIMIT {N} BY ({order_by_cols_str})'
        return sample_clause

    @staticmethod
    def _assemble_limit(limit_n=None) -> str:

        if limit_n is None:
            SAMPLE_CLAUSE = ''
        else:
            SAMPLE_CLAUSE = f'LIMIT {limit_n} '
        return SAMPLE_CLAUSE

    @staticmethod
    def raw_create_select_sql(select_clause: str, db_table: str, sample_clause: str, array_join_clause: str,
                              join_clause: str, prewhere_clause: str, where_clause: str, group_by_clause: str,
                              having_clause: str, order_by_clause: str, limit_n_clause: str, limit_clause: str) -> str:
        """

        :param select_clause:
        :param db_table:
        :param sample_clause:
        :param array_join_clause:
        :param join_clause:
        :param prewhere_clause:
        :param where_clause:
        :param group_by_clause:
        :param having_clause:
        :param order_by_clause:
        :param limit_n_clause:
        :param limit_clause:
        :return:
        """
        """SELECT [DISTINCT] expr_list
                    [FROM [db.]table | (subquery) | table_function] [FINAL]
                    [SAMPLE sample_coeff]
                    [ARRAY JOIN ...]
                    [GLOBAL] ANY|ALL INNER|LEFT JOIN (subquery)|table USING columns_list
                    [PREWHERE expr]
                    [WHERE expr]
                    [GROUP BY expr_list] [WITH TOTALS]
                    [HAVING expr]
                    [ORDER BY expr_list]
                    [LIMIT n BY columns]
                    [LIMIT [n, ]m]
                    [UNION ALL ...]
                    [INTO OUTFILE filename]
                    [FORMAT format]"""
        if db_table.lower().startswith('select '):
            db_table = f"( {db_table} )"
        else:
            pass
        main_body = f"SELECT {select_clause} FROM {db_table} {sample_clause}"
        join = f"{array_join_clause} {join_clause}"
        where_conditions = f"{prewhere_clause} {where_clause} {group_by_clause} {having_clause} "
        order_limit = f"{order_by_clause} {limit_n_clause} {limit_clause}"
        sql = f"{main_body} {join} {where_conditions} {order_limit} SETTINGS joined_subquery_requires_alias=0"
        return sql

    @classmethod
    def select(cls, db_table: str,
               cols: (list, str, None) = None,
               sample: (int, float, None) = None,
               array_join: (list, None) = None,
               join: (dict, None) = None,
               prewhere: (list, None) = None,
               where: (list, None) = None,
               having: (list, None) = None,
               group_by: (list, None) = None,
               order_by: (list, None) = None,
               limit_by: (dict, None) = None,
               limit: (int, None) = None) -> str:
        """

        :param having: str ["r1 >1 and r2 <2"]
        :param db_table: str default.test
        :param cols: list [ r1,r2,r3 ]
        :param sample: str 0.1 or 1000
        :param array_join: list ['arrayA as a','arrayB as b']
        :param join: dict {'type':'all left join','USING' : "r1,r2",'sql':'select * from test'}
        :param prewhere: str ["r1 >1" , "r2 <2"]
        :param where: str ["r1 >1.5" , "r2 <1.3"]
        :param group_by: list ['r1','r2']
        :param order_by: list ['r1 desc','r2 desc']
        :param limit_by: dict {'N':10,'limit_by_cols':['r1','r2']}
        :param limit: int 100
        :return:  str
        """
        if isinstance(cols, str):
            select_clause = cols
        elif isinstance(cols, list):
            select_clause = ','.join(cols)
        elif cols is None:
            select_clause = '*'
        else:
            raise ValueError('cols only accept str or list')
        SAMPLE_CLAUSE = cls._assemble_sample(sample=sample)
        ARRAY_JOIN_CLAUSE = cls._assemble_array_join(array_join_list=array_join)
        JOIN_CLAUSE = cls._assemble_join(join)
        PREWHERE_CLAUSE = cls._assemble_where_like(prewhere, prefix='PREWHERE')
        WHERE_CLAUSE = cls._assemble_where_like(where, prefix='WHERE')
        HAVING_CLAUSE = cls._assemble_where_like(having, prefix='HAVING')
        GROUP_BY_CLAUSE = cls._assemble_group_by(group_by)
        ORDER_BY_CLAUSE = cls._assemble_order_by(order_by)
        LIMIT_N_CLAUSE = cls._assemble_limit_by(limit_by)
        LIMIT_CLAUSE = cls._assemble_limit(limit)

        return cls.raw_create_select_sql(select_clause, db_table, SAMPLE_CLAUSE, ARRAY_JOIN_CLAUSE, JOIN_CLAUSE,
                                         PREWHERE_CLAUSE, WHERE_CLAUSE, GROUP_BY_CLAUSE, HAVING_CLAUSE, ORDER_BY_CLAUSE,
                                         LIMIT_N_CLAUSE, LIMIT_CLAUSE)


class CreateBuilder(object):
    @staticmethod
    def _assemble_conditions_clause(prefix_clause: str, cols: (list, tuple, None), default: str = '') -> str:
        """

        :param prefix_clause:
        :param cols:
        :param default:
        :return:
        """
        if cols is None:
            return default
        else:
            cols_str = ','.join(cols)
            if len(cols) >= 2:
                return f"{prefix_clause} ( {cols_str} ) "
            else:
                return f"{prefix_clause}  {cols_str}  "


class CreateTableFromSQLUtils(object):
    @classmethod
    def _obtain_describe_sql(cls, sql: str, pattern: str = r'[\s]+limit[\s]+[0-9]+$') -> str:
        """
        detect limit end and obtain describe sql
        :param sql:
        :param pattern:
        :return:
        """
        limit_status = cls._check_end_with_limit(sql, pattern=pattern)
        if limit_status:
            describe_sql = f' describe({sql}) '
        else:
            describe_sql = f'describe ( {sql} limit 1)'

        return describe_sql

    @staticmethod
    def _create_table_sql(db: str, table: str, select_or_dtypes_dict: (str, dict),
                          order_by_key_cols: (list, tuple),
                          primary_key_cols: (list, tuple, None) = None,
                          sample_by_cols: (list, tuple, None) = None,
                          partition_by_cols: (list, tuple, None) = None,
                          engine_type: str = 'ReplacingMergeTree',
                          on_cluster: str = '',
                          settings: str = "SETTINGS index_granularity = 8192",
                          other: str = '') -> str:
        """
        CREATE TABLE [IF NOT EXISTS] [db.]table_name [ON CLUSTER cluster]
                   (
                       name1 [type1] [DEFAULT|MATERIALIZED|ALIAS expr1],
                       name2 [type2] [DEFAULT|MATERIALIZED|ALIAS expr2],
                       ...
                   ) ENGINE = ReplacingMergeTree([ver])
                   [PARTITION BY expr]
                   [ORDER BY expr]
                   [PRIMARY KEY expr]
                   [SAMPLE BY expr]
                   [SETTINGS name=value, ...]

        :param db:
        :param table:
        :param select_or_dtypes_dict:
        :param order_by_key_cols:
        :param primary_key_cols:
        :param sample_by_cols:
        :param engine_type:
        :param partition_by_cols:
        :param settings:
        :param other:
        :return:
        """

        default = ""

        # order_by_clause = f"ORDER BY ( {','.join(order_by_key_cols)} )"
        order_by_clause = CreateBuilder._assemble_conditions_clause('ORDER BY', order_by_key_cols, default=default)
        sample_clause = CreateBuilder._assemble_conditions_clause('SAMPLE BY', sample_by_cols, default=default)
        primary_by_clause = CreateBuilder._assemble_conditions_clause('PRIMARY KEY ', primary_key_cols, default=default)
        partition_by_clause = CreateBuilder._assemble_conditions_clause('PARTITION BY', partition_by_cols,
                                                                        default=default)

        # if partitions_expr is None:
        #     partition_by_clause = ''
        # else:
        #     partition_by_clause = f"PARTITION BY {partitions_expr}"

        # if sample_expr is None:
        #     sample_clause = ''
        # else:
        #     sample_clause = sample_expr

        # if primary_key_cols is None:
        #     primary_by_clause = ''
        # else:
        #     primary_key_expr = ','.join(primary_key_cols)
        #     primary_by_clause = f"PRIMARY BY ({primary_key_expr})"
        if isinstance(select_or_dtypes_dict, dict):
            cols_def = ','.join([f"`{name}` {d_type}" for name, d_type in select_or_dtypes_dict.items()])
            main_body = f"CREATE TABLE IF NOT EXISTS {db}.{table} {on_cluster} ( {cols_def} ) ENGINE = {engine_type}"
            tail = ''
        elif isinstance(select_or_dtypes_dict, str) and select_or_dtypes_dict.lower().startswith('select'):
            main_body = f"CREATE TABLE IF NOT EXISTS {db}.{table} {on_cluster}  ENGINE = {engine_type} "
            tail = f'as {select_or_dtypes_dict}'
        else:
            raise ParameterTypeError(
                f'select_or_dtypes_dict only accept dict or str start with select! but get {select_or_dtypes_dict}')

        cond_clause = f" {partition_by_clause} {order_by_clause} {primary_by_clause} {sample_clause} "

        base = f"{main_body} {cond_clause}  {other} {settings} {tail} "

        return base

    @staticmethod
    def _check_end_with_limit(string: str, pattern: str = r'[\s]+limit[\s]+[0-9]+$'):
        """

        :param string:
        :param pattern:
        :return:
        """

        m = re.findall(pattern, string)
        if m is None or m == []:
            return False
        else:
            return True

    @staticmethod
    def _translate_dtypes1_as_dtypes2(df: pd.DataFrame, src2target=None):
        """

        :param df:
        :param src2target:
        :return:
        """
        if src2target is None:
            src2target = {'category': 'str'}
        dtypes_series = df.dtypes
        for src, dest in src2target.items():
            if src in dtypes_series:
                category_cols = dtypes_series[dtypes_series == src].index
                for col in category_cols:
                    df[col] = df[col].astype(dest)
            else:
                pass
        return df

    @staticmethod
    def _translate_dtypes_from_df(df: pd.DataFrame, translate_dtypes=None):
        """

        :param df:
        :param translate_dtypes:
        :return:
        """

        if translate_dtypes is None:
            translate_dtypes = {'object': 'String',
                                'datetime64[ns]': 'Datetime'}
        if hasattr(df, 'dtypes'):
            dtypes_series = df.dtypes.replace(translate_dtypes)
            return dtypes_series.map(lambda x: str(x).capitalize()).to_dict()
        elif hasattr(df, '_columns_') and 'type' in df._columns_ and 'name' in df._columns_:
            dtypes_series = df.set_index('name')['type'].replace(translate_dtypes)
            return dtypes_series.map(lambda x: str(x)).to_dict()
        else:
            raise ValueError(f'unknown df:{type(df)}')

    @classmethod
    def _create_table_from_df(cls, db: str, table: str, df: pd.DataFrame, key_cols: (list, tuple),
                              primary_key_cols: (list, tuple, None) = None,
                              sample_by_cols: (list, tuple, None) = None,
                              partition_by_cols: (list, tuple, None) = None,
                              on_cluster: str = '',
                              engine_type: str = 'ReplacingMergeTree',
                              extra_format_dict: dict = None,
                              settings: str = "SETTINGS index_granularity = 8192",
                              other: str = ''):
        """

        :param db:  str
        :param table:
        :param df:
        :param key_cols:
        :param primary_key_cols:
        :param sample_by_cols:
        :param partition_by_cols:
        :param on_cluster:
        :param engine_type:
        :param extra_format_dict:
        :param settings:
        :param other:
        :return:
        """

        df = cls._translate_dtypes1_as_dtypes2(df, src2target={'category': 'str'})
        cols = df.columns
        dtypes_dict = cls._translate_dtypes_from_df(df)
        if extra_format_dict is None:
            pass
        else:
            dtypes_dict._update(extra_format_dict)
        dtypes_str_dict = {k: v for k, v in dtypes_dict.items() if k in cols}
        base = cls._create_table_sql(db, table, dtypes_str_dict, key_cols,
                                     primary_key_cols=primary_key_cols,
                                     sample_by_cols=sample_by_cols,
                                     engine_type=engine_type,
                                     on_cluster=on_cluster,
                                     partition_by_cols=partition_by_cols,
                                     settings=settings,
                                     other=other)
        return base

    @classmethod
    def _create_table_from_sql(cls, db: str, table: str, sql: str, key_cols: list,
                               extra_format_dict: (dict, None) = None,
                               primary_key_cols: (list, tuple, None) = None,
                               sample_by_cols: (list, tuple, None) = None,
                               partition_by_cols: (list, tuple, None) = None,
                               engine_type: str = 'ReplacingMergeTree',
                               on_cluster: str = '',
                               settings: str = "SETTINGS index_granularity = 8192",
                               other: str = '',
                               ):
        """


        :param db:
        :param table:
        :param sql:
        :param key_cols:
        :param engine_type:
        :param extra_format_dict:
        :return:
        """

        if isinstance(sql, str):
            # if self._check_sql_type(sql) == 'insert-liked':
            if sql.lower().startswith('create'):
                # if sql is create table sql will return directly
                return sql
            elif sql.lower().startswith('select'):
                pass
            else:
                raise ParameterTypeError(f'sql must be string and start with select! but get {sql}')
        else:
            raise ParameterTypeError(f'sql must be string and start with select! but get {sql}')
        # detect end with limit
        # describe_sql = cls._obtain_describe_sql(sql)
        # dtypes_df = query(describe_sql)
        # dtypes_dict = dict(dtypes_df[['name', 'type']].drop_duplicates().values)

        # if isinstance(extra_format_dict, dict):
        #     dtypes_dict.update(extra_format_dict)
        # else:
        #     pass
        sql = cls._create_table_sql(db, table, sql, key_cols,
                                    primary_key_cols=primary_key_cols,
                                    sample_by_cols=sample_by_cols, on_cluster=on_cluster,
                                    engine_type=engine_type,
                                    partition_by_cols=partition_by_cols,
                                    settings=settings,
                                    other=other)

        # self.query(sql)
        return sql

    @classmethod
    def creator_sql(cls, db: str, table: str, df_or_sql: (pd.DataFrame, str), key_cols: (list, tuple),
                    extra_format_dict: (dict, None) = None,
                    primary_key_cols: (list, tuple, None) = None,
                    sample_by_cols: (list, tuple, None) = None,
                    partition_by_cols: (list, tuple, None) = None,
                    on_cluster='',
                    engine_type: str = 'ReplacingMergeTree',
                    settings: str = "SETTINGS index_granularity = 8192",
                    other: str = '',
                    ):
        """

        :param db:
        :param table:
        :param df_or_sql:
        :param key_cols:
        :param extra_format_dict:
        :param primary_key_cols:
        :param sample_by_cols:
        :param partition_by_cols:
        :param on_cluster:
        :param engine_type:
        :param settings:
        :param other:
        :return:
        """
        if isinstance(df_or_sql, pd.DataFrame):
            return cls._create_table_from_df(db, table, df_or_sql, key_cols,
                                             primary_key_cols=primary_key_cols, sample_by_cols=sample_by_cols,
                                             engine_type=engine_type, extra_format_dict=extra_format_dict,
                                             partition_by_cols=partition_by_cols, on_cluster=on_cluster,
                                             settings=settings,
                                             other=other)
        else:
            return cls._create_table_from_sql(db, table, df_or_sql, key_cols, on_cluster=on_cluster,
                                              primary_key_cols=primary_key_cols, sample_by_cols=sample_by_cols,
                                              engine_type=engine_type, extra_format_dict=extra_format_dict,
                                              partition_by_cols=partition_by_cols,
                                              settings=settings,
                                              other=other)
        pass


class CreateTableFromInfoUtils(object):

    @staticmethod
    def _raw_create_table_sql(db_table: str,
                              cols_def: str,
                              order_by_clause: str,
                              primary_by_clause: str = '',
                              sample_clause: str = '',
                              engine_type: str = 'ReplacingMergeTree',
                              on_cluster: str = '',
                              partition_by_clause: str = '',
                              ttl: str = '',
                              settings: str = "SETTINGS index_granularity = 8192"):
        # TODO add ttl expr at future
        """
        CREATE TABLE [IF NOT EXISTS] [db.]table_name [ON CLUSTER cluster]
                    (
                        name1 [type1] [DEFAULT|MATERIALIZED|ALIAS expr1],
                        name2 [type2] [DEFAULT|MATERIALIZED|ALIAS expr2],
                        ...
                    ) ENGINE = ReplacingMergeTree([ver])
                    [PARTITION BY expr]
                    [ORDER BY expr]
                    [PRIMARY KEY expr]
                    [SAMPLE BY expr]
                    [SETTINGS name=value, ...]
        :param ttl:
        :param on_cluster:
        :param sample_clause:
        :param primary_by_clause:
        :param partition_by_clause:
        :param db_table:
        :param cols_def:
        :param order_by_clause:
        :param engine_type:
        :return:
        """

        maid_body = f"CREATE TABLE IF NOT EXISTS {db_table} {on_cluster} ( {cols_def} ) ENGINE = {engine_type}"

        conds = f"{partition_by_clause} {order_by_clause} {primary_by_clause} {sample_clause} {ttl}"

        base = f"{maid_body} {conds}  {settings}"
        return base

    @classmethod
    def creator_info(cls,
                     db_table: str,
                     var_dict: dict,
                     order_by_cols: (list, tuple),
                     sample_by_cols: (list, tuple, None) = None,
                     partition_by_cols: (list, tuple, None) = None,
                     primary_by_cols: (list, tuple, None) = None,
                     on_cluster: str = '',
                     engine_type: str = 'ReplacingMergeTree',
                     ttl: str = '',
                     settings="SETTINGS index_granularity = 8192"):
        """

        :param settings:
        :param engine_type:
        :param var_dict:
        :param ttl:
        :param db_table:
        :param order_by_cols:
        :param sample_by_cols:
        :param on_cluster:
        :param partition_by_cols:
        :param primary_by_cols:
        :return:
        """
        default = ''
        cols_def = ','.join([f"`{var}` {v}" for var, v in var_dict.items()])

        order_by_clause = CreateBuilder._assemble_conditions_clause('ORDER BY', order_by_cols, default=default)
        sample_clause = CreateBuilder._assemble_conditions_clause('SAMPLE BY', sample_by_cols, default=default)
        primary_by_clause = CreateBuilder._assemble_conditions_clause('PRIMARY KEY', primary_by_cols, default=default)
        partition_by_clause = CreateBuilder._assemble_conditions_clause('PARTITION BY', partition_by_cols,
                                                                        default=default)

        return cls._raw_create_table_sql(db_table, cols_def, order_by_clause,
                                         primary_by_clause=primary_by_clause,
                                         sample_clause=sample_clause,
                                         engine_type=engine_type, on_cluster=on_cluster,
                                         partition_by_clause=partition_by_clause, ttl=ttl, settings=settings)


class CreateTableUtils(CreateTableFromInfoUtils, CreateTableFromSQLUtils):
    # TODO test merge function for create table
    #
    @classmethod
    def _create(cls, db: str, table: str, df_or_sql_or_dict: (pd.DataFrame, str, dict),
                key_cols: (list, tuple),
                extra_format_dict: (dict, None) = None,
                primary_key_cols: (list, tuple, None) = None,
                sample_by_cols: (list, tuple, None) = None,
                partition_by_cols: (list, tuple, None) = None,
                settings: str = "SETTINGS index_granularity = 8192",
                engine_type: str = 'ReplacingMergeTree',
                on_cluster: str = '',
                ttl: str = '',
                other: str = ''):
        """


        df_or_sql_or_dict

        :param db:
        :param table:
        :param df_or_sql_or_dict: pd.DataFrame & sql string or dtype dict
        :param key_cols:
        :param extra_format_dict:
        :param primary_key_cols:
        :param sample_by_cols:
        :param partition_by_cols:
        :param settings:
        :param engine_type:
        :param on_cluster:
        :param ttl:
        :param other:
        :return:
        """
        if isinstance(df_or_sql_or_dict, dict):
            db_table = f"{db}.{table}"
            var_dict = df_or_sql_or_dict
            sql = cls.creator_info(db_table, var_dict,
                                   order_by_cols=key_cols,
                                   sample_by_cols=sample_by_cols,  #: (list, tuple, None)
                                   partition_by_cols=partition_by_cols,  # : (list, tuple, None)
                                   primary_by_cols=primary_key_cols,  # : (list, tuple, None)
                                   on_cluster=on_cluster,
                                   engine_type=engine_type,
                                   ttl=ttl, settings=settings)
        elif isinstance(df_or_sql_or_dict, (pd.DataFrame, str)):

            sql = cls.creator_sql(db, table, df_or_sql_or_dict, key_cols,
                                  extra_format_dict=extra_format_dict,
                                  sample_by_cols=sample_by_cols,  #: (list, tuple, None)
                                  partition_by_cols=partition_by_cols,  # : (list, tuple, None)
                                  primary_key_cols=primary_key_cols,  # : (list, tuple, None)
                                  on_cluster=on_cluster,
                                  engine_type=engine_type,
                                  settings=settings,
                                  other=ttl + ' ' + other)

        else:
            raise ValueError(f'unsupported type of df_or_sql_or_dict: {type(df_or_sql_or_dict)}')
        return sql


class TableEngineCreator(ClickHouseTableNode, CreateTableUtils, SQLBuilder):
    def __init__(self, conn_str: (str, None) = None, **kwarg):
        super(TableEngineCreator, self).__init__(conn_str=conn_str, **kwarg)

    def _create_view(self, db: str, table: str, sql: str,
                     key_cols: list,
                     primary_key_cols=None,
                     sample_by_cols=None,
                     engine_type: str = 'view',
                     extra_format_dict: (dict, None) = None,
                     partition_by_cols: (str, None) = None,
                     **kwargs):
        """

        :param db:
        :param table:
        :param sql:
        :param key_cols:
        :param primary_key_cols:
        :param sample_by_cols:
        :param engine_type:
        :param extra_format_dict:
        :param partition_by_cols:
        :param kwargs:
        :return:
        """
        if not self._check_exists(f'{db}.{table}', mode='table'):
            pass
        else:
            raise ClickHouseTableExistsError(f'{db}.{table} is exists!')
        if engine_type.lower() == 'view':
            create_sql = f'create view if not exists {db}.{table} as {sql}'
            self.query(create_sql)
        else:
            raise ValueError(f'current not support material view : {engine_type}')

    # def create_table_if_not_exists(self, db: str, table: str, sql: (str, pd.DataFrame, None),
    #                                key_cols: list,
    #                                primary_key_cols=None, sample_expr=None,
    #                                engine_type: str = 'ReplacingMergeTree',
    #                                extra_format_dict: (dict, None) = None, partitions_expr: (str, None) = None):
    #     """
    #
    #     :param db:
    #     :param table:
    #     :param sql:
    #     :param key_cols:
    #     :param primary_key_cols:
    #     :param sample_expr:
    #     :param engine_type:
    #     :param extra_format_dict:
    #     :param partitions_expr:
    #     :return:
    #     """
    #
    #     self._create_table(db, table, sql, key_cols,
    #                            primary_key_cols=primary_key_cols,
    #                            sample_expr=sample_expr,
    #                            engine_type=engine_type,
    #                            extra_format_dict=extra_format_dict,
    #                            partitions_expr=partitions_expr)

    def create(self,
               db: str,
               table: str,
               df_or_sql_or_dict: (str, pd.DataFrame, dict),
               key_cols: list,
               engine_type: str = 'ReplacingMergeTree',
               primary_key_cols: (list, tuple, None) = None,
               sample_by_cols: (list, tuple, None) = None,
               extra_format_dict: (dict, None) = None,
               partition_by_cols: (list, tuple, None) = None,
               settings: str = "SETTINGS index_granularity = 8192",
               on_cluster: str = '',
               ttl: str = '',
               other: str = '',
               check: bool = True,
               execute: bool = False):
        """

        df_or_sql_or_dict : can be set up sql query or data or create table dtypes dict

        :param db:
        :param table:
        :param df_or_sql_or_dict:
        :param key_cols:
        :param engine_type:
        :param primary_key_cols:
        :param sample_by_cols:
        :param extra_format_dict:
        :param partition_by_cols:
        :param settings:
        :param on_cluster:
        :param ttl:
        :param other:
        :param check:
        :param execute:
        :return:
        """

        # if self._check_exists(f'{db}.{table}', mode='table'):
        #     raise ClickHouseTableExistsError(f'{db}.{table} is exists!')
        # else:
        create_sql = self._create(db, table, df_or_sql_or_dict,
                                  key_cols,
                                  extra_format_dict=extra_format_dict,
                                  primary_key_cols=primary_key_cols,
                                  sample_by_cols=sample_by_cols,
                                  partition_by_cols=partition_by_cols,
                                  settings=settings,
                                  engine_type=engine_type,
                                  on_cluster=on_cluster,
                                  ttl=ttl,
                                  other=other)
        if check:
            if self._check_exists(db_table=f'{db}.{table}'):
                raise ClickHouseTableExistsError(f'{db}.{table} is exists!')
        if execute:
            self.query(create_sql)
        else:
            return create_sql


if __name__ == '__main__':
    pass



# coding=utf-8


class MergeSQLUtils(object):
    # merge table
    @staticmethod
    def _merge(first,
               seconds: str,
               using: (list, str, tuple),
               cols: (list, str, None) = None,
               join_type='all full join',
               # cols: list,
               #  sample: (int, float, None) = None,
               #  array_join: (list, None) = None,
               #  join: (dict, None) = None,
               #  prewhere: (list, None) = None,
               #  where: (list, None) = None,
               #  having: (list, None) = None,
               #  group_by: (list, None) = None,
               #  order_by: (list, None) = None,
               #  limit_by: (dict, None) = None,
               #  limit: (int, None) = None
               ) -> str:
        from ClickSQL.clickhouse.ClickHouseCreate import SQLBuilder
        # self._complex = True
        if isinstance(using, (list, tuple)):
            using = ','.join(using)

        join = {'type': join_type, 'USING': using, 'sql': str(seconds)}
        sql = SQLBuilder.select(str(first), cols, join=join, limit=None)
        # if execute:
        #     return self.operator(sql)
        # else:
        return sql


if __name__ == '__main__':
    pass



# coding=utf-8
def chunk(obj, chunks=2000):
    if hasattr(obj, '__len__'):
        length = len(obj)
        for i in range(0, length, chunks):
            yield obj[i:i + chunks]


if __name__ == '__main__':
    pass



# coding=utf-8
import os

from setuptools import setup, find_packages

from ClickSQL import __version__, __author__

# 读取文件内容
this_directory = os.path.abspath(os.path.dirname(__file__))


def read_file(filename):
    with open(os.path.join(this_directory, filename), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


# 获取依赖
def read_requirements(filename):
    return [line.strip() for line in read_file(filename).splitlines()
            if not line.startswith('#')]


setup(
    name="ClickSQL",
    version=__version__,
    # python_requires='>=3.6',  # python环境
    keywords=("ClickSQL", "Databases"),
    description="SQL programming",
    long_description=read_file('README.md'),
    long_description_content_type="text/markdown",  # 新参数
    license="MIT Licence",

    url="http://www.github.com/sn0wfree/ClickSQL",
    author=__author__,
    author_email="snowfreedom0815@gmail.com",

    packages=find_packages(exclude=['dist', 'docs', 'ClickSQL.egg-info']),
    include_package_data=True,
    setup_requires=read_requirements('requirements.txt'),  # 指定需要安装的依赖,

)
